"""add column asset body.

Revision ID: 7cba9eeaaf1d
Revises: f0367d611631
Create Date: 2021-03-17 16:40:52.449012

"""

from dagster._core.storage.migration.utils import add_asset_details_column

# revision identifiers, used by Alembic.
revision = "7cba9eeaaf1d"
down_revision = "f0367d611631"
branch_labels = None
depends_on = None


def upgrade():
    add_asset_details_column()


def downgrade():
    pass



import contextlib
import inspect
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterator, Mapping
from typing import (  # noqa: UP035
    AbstractSet,
    Any,
    Callable,
    Generic,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseModel
from typing_extensions import TypeAlias, TypeGuard, get_args, get_origin

import dagster._check as check
from dagster import Field as DagsterField
from dagster._annotations import deprecated
from dagster._config.field_utils import config_dictionary_from_values
from dagster._config.pythonic_config.attach_other_object_to_context import (
    IAttachDifferentObjectToOpContext as IAttachDifferentObjectToOpContext,
)
from dagster._config.pythonic_config.config import (
    Config,
    MakeConfigCacheable,
    infer_schema_from_config_class,
)
from dagster._config.pythonic_config.conversion_utils import TResValue, _curry_config_schema
from dagster._config.pythonic_config.typing_utils import (
    BaseResourceMeta,
    LateBoundTypesForResourceTypeChecking,
    TypecheckAllowPartialResourceInitParams,
)
from dagster._config.validate import validate_config
from dagster._core.decorator_utils import get_function_params
from dagster._core.definitions.definition_config_schema import (
    ConfiguredDefinitionConfigSchema,
    DefinitionConfigSchema,
)
from dagster._core.definitions.resource_definition import (
    ResourceDefinition,
    ResourceFunction,
    ResourceFunctionWithContext,
    ResourceFunctionWithoutContext,
    has_at_least_one_parameter,
)
from dagster._core.definitions.resource_requirement import ResourceRequirement
from dagster._core.errors import DagsterInvalidConfigError, DagsterInvalidDefinitionError
from dagster._core.execution.context.init import InitResourceContext, build_init_resource_context
from dagster._model.pydantic_compat_layer import model_fields
from dagster._record import record
from dagster._utils.cached_method import cached_method
from dagster._utils.typing_api import is_closed_python_optional_type

T_Self = TypeVar("T_Self", bound="ConfigurableResourceFactory")
ResourceId: TypeAlias = int


class NestedResourcesResourceDefinition(ResourceDefinition, ABC):
    @property
    @abstractmethod
    def nested_partial_resources(self) -> Mapping[str, "CoercibleToResource"]: ...

    @property
    @abstractmethod
    def nested_resources(self) -> Mapping[str, Any]: ...

    @property
    @abstractmethod
    def configurable_resource_cls(self) -> type: ...

    def get_resource_requirements(self, source_key: str) -> Iterator["ResourceRequirement"]:
        for attr_name, partial_resource in self.nested_partial_resources.items():
            yield PartialResourceDependencyRequirement(
                class_name=self.configurable_resource_cls.__name__,
                attr_name=attr_name,
                partial_resource=partial_resource,
            )

        for inner_key, resource in self.nested_resources.items():
            if is_coercible_to_resource(resource):
                yield from coerce_to_resource(resource).get_resource_requirements(inner_key)

        yield from super().get_resource_requirements(source_key)

    def get_required_resource_keys(
        self, resource_defs: Mapping[str, ResourceDefinition]
    ) -> AbstractSet[str]:
        resolved_keys = set(self.required_resource_keys)
        for attr_name, partial_resource in self.nested_partial_resources.items():
            if is_coercible_to_resource(partial_resource):
                resolved_keys.add(
                    _resolve_partial_resource_to_key(attr_name, partial_resource, resource_defs)
                )
            else:
                check.failed(
                    f"Unexpected nested partial resource of type {type(partial_resource)} is not coercible to resource"
                )

        for resource in self.nested_resources.values():
            if is_coercible_to_resource(resource):
                resolved_keys.update(
                    coerce_to_resource(resource).get_required_resource_keys(resource_defs)
                )

        return resolved_keys


class ConfigurableResourceFactoryResourceDefinition(NestedResourcesResourceDefinition):
    def __init__(
        self,
        configurable_resource_cls: type,
        resource_fn: ResourceFunction,
        config_schema: Any,
        description: Optional[str],
        nested_resources: Mapping[str, Any],
        nested_partial_resources: Mapping[str, Any],
        dagster_maintained: bool = False,
    ):
        super().__init__(
            resource_fn=resource_fn,
            config_schema=config_schema,
            description=description,
        )
        self._configurable_resource_cls = configurable_resource_cls
        self._nested_partial_resources = nested_partial_resources
        self._nested_resources = nested_resources
        self._dagster_maintained = dagster_maintained

    @property
    def configurable_resource_cls(self) -> type:
        return self._configurable_resource_cls

    @property
    def nested_resources(
        self,
    ) -> Mapping[str, Any]:
        return self._nested_resources

    @property
    def nested_partial_resources(
        self,
    ) -> Mapping[str, "CoercibleToResource"]:
        return self._nested_partial_resources

    def _is_dagster_maintained(self) -> bool:
        return self._dagster_maintained


class ConfigurableResourceFactoryState(NamedTuple):
    nested_partial_resources: Mapping[str, Any]
    resolved_config_dict: dict[str, Any]
    config_schema: DefinitionConfigSchema
    schema: DagsterField
    nested_resources: dict[str, Any]
    resource_context: Optional[InitResourceContext]


class ConfigurableResourceFactory(
    Config,
    TypecheckAllowPartialResourceInitParams,
    Generic[TResValue],
    ABC,
    metaclass=BaseResourceMeta,
):
    """Base class for creating and managing the lifecycle of Dagster resources that utilize structured config.

    Users should directly inherit from this class when they want the object passed to user-defined
    code (such as an asset or op) to be different than the object that defines the configuration
    schema and is passed to the :py:class:`Definitions` object. Cases where this is useful include is
    when the object passed to user code is:

    * An existing class from a third-party library that the user does not control.
    * A complex class that requires substantial internal state management or itself requires arguments beyond its config values.
    * A class with expensive initialization that should not be invoked on code location load, but rather lazily on first use in an op or asset during a run.
    * A class that you desire to be a plain Python class, rather than a Pydantic class, for whatever reason.

    This class is a subclass of both :py:class:`ResourceDefinition` and :py:class:`Config`, and
    must implement ``create_resource``, which creates the resource to pass to user code.

    Example definition:

    .. code-block:: python

        class DatabaseResource(ConfigurableResourceFactory[Database]):
            connection_uri: str

            def create_resource(self, _init_context) -> Database:
                # For example Database could be from a third-party library or require expensive setup.
                # Or you could just prefer to separate the concerns of configuration and runtime representation
                return Database(self.connection_uri)

    To use a resource created by a factory in a job, you must use the Resource type annotation.

    Example usage:


    .. code-block:: python

        @asset
        def asset_that_uses_database(database: ResourceParam[Database]):
            # Database used directly in user code
            database.query("SELECT * FROM table")

        defs = Definitions(
            assets=[asset_that_uses_database],
            resources={"database": DatabaseResource(connection_uri="some_uri")},
        )

    """

    def __init__(self, **data: Any):
        resource_pointers, data_without_resources = separate_resource_params(self.__class__, data)

        schema = infer_schema_from_config_class(
            self.__class__, fields_to_omit=set(resource_pointers.keys())
        )

        # Populate config values
        super().__init__(**data_without_resources, **resource_pointers)

        # We pull the values from the Pydantic config object, which may cast values
        # to the correct type under the hood - useful in particular for enums
        casted_data_without_resources = {
            k: v
            for k, v in self._convert_to_config_dictionary().items()
            if k in data_without_resources
        }
        resolved_config_dict = config_dictionary_from_values(casted_data_without_resources, schema)

        self._state__internal__ = ConfigurableResourceFactoryState(
            # We keep track of any resources we depend on which are not fully configured
            # so that we can retrieve them at runtime
            nested_partial_resources={
                k: v for k, v in resource_pointers.items() if (not _is_fully_configured(v))
            },
            resolved_config_dict=resolved_config_dict,
            # These are unfortunately named very similarily
            config_schema=_curry_config_schema(schema, resolved_config_dict),
            schema=schema,
            nested_resources={k: v for k, v in resource_pointers.items()},
            resource_context=None,
        )

    @property
    def _schema(self):
        return self._state__internal__.schema

    @property
    def _config_schema(self):
        return self._state__internal__.config_schema

    @property
    def _nested_partial_resources(self):
        return self._state__internal__.nested_partial_resources

    @property
    def _nested_resources(self):
        return self._state__internal__.nested_resources

    @property
    def _resolved_config_dict(self):
        return self._state__internal__.resolved_config_dict

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        """This should be overridden to return True by all dagster maintained resources and IO managers."""
        return False

    @classmethod
    def _is_cm_resource_cls(cls: type["ConfigurableResourceFactory"]) -> bool:
        return (
            cls.yield_for_execution != ConfigurableResourceFactory.yield_for_execution
            or cls.teardown_after_execution != ConfigurableResourceFactory.teardown_after_execution
            # We assume that any resource which has nested resources needs to be treated as a
            # context manager resource, since its nested resources may be context managers
            # and need setup and teardown logic
            or len(_get_resource_param_fields(cls)) > 0
        )

    @property
    def _is_cm_resource(self) -> bool:
        return self.__class__._is_cm_resource_cls()  # noqa: SLF001

    def _get_initialize_and_run_fn(self) -> Callable:
        return self._initialize_and_run_cm if self._is_cm_resource else self._initialize_and_run

    @cached_method  # resource resolution depends on always resolving to the same ResourceDefinition instance
    def get_resource_definition(self) -> ConfigurableResourceFactoryResourceDefinition:
        return ConfigurableResourceFactoryResourceDefinition(
            self.__class__,
            resource_fn=self._get_initialize_and_run_fn(),
            config_schema=self._config_schema,
            description=self.__doc__,
            nested_resources=self.nested_resources,
            nested_partial_resources=self._nested_partial_resources,
            dagster_maintained=self._is_dagster_maintained(),
        )

    @abstractmethod
    def create_resource(self, context: InitResourceContext) -> TResValue:
        """Returns the object that this resource hands to user code, accessible by ops or assets
        through the context or resource parameters. This works like the function decorated
        with @resource when using function-based resources.
        """
        raise NotImplementedError()

    @property
    def nested_resources(
        self,
    ) -> Mapping[str, Any]:
        return self._nested_resources

    @classmethod
    def configure_at_launch(cls: "type[T_Self]", **kwargs) -> "PartialResource[T_Self]":
        """Returns a partially initialized copy of the resource, with remaining config fields
        set at runtime.
        """
        return PartialResource(cls, data=kwargs)

    def _with_updated_values(
        self, values: Optional[Mapping[str, Any]]
    ) -> "ConfigurableResourceFactory[TResValue]":
        """Returns a new instance of the resource with the given values.
        Used when initializing a resource at runtime.
        """
        values = check.opt_mapping_param(values, "values", key_type=str)
        # Since Resource extends BaseModel and is a dataclass, we know that the
        # signature of any __init__ method will always consist of the fields
        # of this class. We can therefore safely pass in the values as kwargs.
        to_populate = self.__class__._get_non_default_public_field_values_cls(  # noqa: SLF001
            {**self._get_non_default_public_field_values(), **values}
        )
        out = self.__class__(**to_populate)
        out._state__internal__ = out._state__internal__._replace(  # noqa: SLF001
            resource_context=self._state__internal__.resource_context
        )
        return out

    @contextlib.contextmanager
    def _resolve_and_update_nested_resources(
        self, context: InitResourceContext
    ) -> Generator["ConfigurableResourceFactory[TResValue]", None, None]:
        """Updates any nested resources with the resource values from the context.
        In this case, populating partially configured resources or
        resources that return plain Python types.

        Returns a new instance of the resource.
        """
        from dagster._core.execution.build_resources import wrap_resource_for_execution

        partial_resources_to_update: dict[str, Any] = {}
        if self._nested_partial_resources:
            for attr_name, resource in self._nested_partial_resources.items():
                key = _resolve_partial_resource_to_key(
                    attr_name, resource, context.all_resource_defs
                )
                resolved_resource = getattr(context.resources, key)
                partial_resources_to_update[attr_name] = resolved_resource

        # Also evaluate any resources that are not partial
        with contextlib.ExitStack() as stack:
            resources_to_update, _ = separate_resource_params(self.__class__, self.__dict__)
            resources_to_update = {
                attr_name: _call_resource_fn_with_default(
                    stack, wrap_resource_for_execution(resource), context
                )
                for attr_name, resource in resources_to_update.items()
                if attr_name not in partial_resources_to_update
            }

            to_update = {**resources_to_update, **partial_resources_to_update}
            yield self._with_updated_values(to_update)

    @deprecated(
        breaking_version="2.0", additional_warn_text="Use `with_replaced_resource_context` instead"
    )
    def with_resource_context(
        self, resource_context: InitResourceContext
    ) -> "ConfigurableResourceFactory[TResValue]":
        return self.with_replaced_resource_context(resource_context)

    def with_replaced_resource_context(
        self, resource_context: InitResourceContext
    ) -> "ConfigurableResourceFactory[TResValue]":
        """Returns a new instance of the resource with the given resource init context bound."""
        # This utility is used to create a copy of this resource, without adjusting
        # any values in this case
        copy = self._with_updated_values({})
        copy._state__internal__ = copy._state__internal__._replace(  # noqa: SLF001
            resource_context=resource_context
        )
        return copy

    def _initialize_and_run(self, context: InitResourceContext) -> TResValue:
        with self._resolve_and_update_nested_resources(context) as has_nested_resource:
            updated_resource = has_nested_resource.with_replaced_resource_context(  # noqa: SLF001
                context
            )._with_updated_values(context.resource_config)

            updated_resource.setup_for_execution(context)
            return updated_resource.create_resource(context)

    @contextlib.contextmanager
    def _initialize_and_run_cm(
        self, context: InitResourceContext
    ) -> Generator[TResValue, None, None]:
        with self._resolve_and_update_nested_resources(context) as has_nested_resource:
            updated_resource = has_nested_resource.with_replaced_resource_context(  # noqa: SLF001
                context
            )._with_updated_values(context.resource_config)

            with updated_resource.yield_for_execution(context) as value:
                yield value

    def setup_for_execution(self, context: InitResourceContext) -> None:
        """Optionally override this method to perform any pre-execution steps
        needed before the resource is used in execution.
        """
        pass

    def teardown_after_execution(self, context: InitResourceContext) -> None:
        """Optionally override this method to perform any post-execution steps
        needed after the resource is used in execution.

        teardown_after_execution will be called even if any part of the run fails.
        It will not be called if setup_for_execution fails.
        """
        pass

    @contextlib.contextmanager
    def yield_for_execution(self, context: InitResourceContext) -> Generator[TResValue, None, None]:
        """Optionally override this method to perform any lifecycle steps
        before or after the resource is used in execution. By default, calls
        setup_for_execution before yielding, and teardown_after_execution after yielding.

        Note that if you override this method and want setup_for_execution or
        teardown_after_execution to be called, you must invoke them yourself.
        """
        self.setup_for_execution(context)
        try:
            yield self.create_resource(context)
        finally:
            self.teardown_after_execution(context)

    def get_resource_context(self) -> InitResourceContext:
        """Returns the context that this resource was initialized with."""
        return check.not_none(
            self._state__internal__.resource_context,
            additional_message="Attempted to get context before resource was initialized.",
        )

    def process_config_and_initialize(self) -> TResValue:
        """Initializes this resource, fully processing its config and returning the prepared
        resource value.
        """
        from dagster._config.post_process import post_process_config

        return self.from_resource_context(
            build_init_resource_context(
                config=post_process_config(
                    self._config_schema.config_type,  # pyright: ignore[reportArgumentType]
                    self._convert_to_config_dictionary(),  # pyright: ignore[reportArgumentType]
                ).value,
            ),
            nested_resources=self.nested_resources,
        )

    @contextlib.contextmanager
    def process_config_and_initialize_cm(self) -> Generator[TResValue, None, None]:
        """Context which initializes this resource, fully processing its config and yielding the
        prepared resource value.
        """
        from dagster._config.post_process import post_process_config

        with self.from_resource_context_cm(
            build_init_resource_context(
                config=post_process_config(
                    self._config_schema.config_type,  # pyright: ignore[reportArgumentType]
                    self._convert_to_config_dictionary(),  # pyright: ignore[reportArgumentType]
                ).value
            ),
            nested_resources=self.nested_resources,
        ) as out:
            yield out

    @classmethod
    def from_resource_context(
        cls, context: InitResourceContext, nested_resources: Optional[Mapping[str, Any]] = None
    ) -> TResValue:
        """Creates a new instance of this resource from a populated InitResourceContext.
        Useful when creating a resource from a function-based resource, for backwards
        compatibility purposes.

        For resources that have custom teardown behavior, use from_resource_context_cm instead.

        Example usage:

        .. code-block:: python

            class MyResource(ConfigurableResource):
                my_str: str

            @resource(config_schema=MyResource.to_config_schema())
            def my_resource(context: InitResourceContext) -> MyResource:
                return MyResource.from_resource_context(context)

        """
        check.invariant(
            not cls._is_cm_resource_cls(),
            "Use from_resource_context_cm for resources which have custom teardown behavior,"
            " e.g. overriding yield_for_execution or teardown_after_execution",
        )
        return cls(  # noqa: SLF001
            **{**(context.resource_config or {}), **(nested_resources or {})}
        )._initialize_and_run(context)

    @classmethod
    @contextlib.contextmanager
    def from_resource_context_cm(
        cls, context: InitResourceContext, nested_resources: Optional[Mapping[str, Any]] = None
    ) -> Generator[TResValue, None, None]:
        """Context which generates a new instance of this resource from a populated InitResourceContext.
        Useful when creating a resource from a function-based resource, for backwards
        compatibility purposes. Handles custom teardown behavior.

        Example usage:

        .. code-block:: python

            class MyResource(ConfigurableResource):
                my_str: str

            @resource(config_schema=MyResource.to_config_schema())
            def my_resource(context: InitResourceContext) -> Generator[MyResource, None, None]:
                with MyResource.from_resource_context_cm(context) as my_resource:
                    yield my_resource

        """
        with cls(  # noqa: SLF001
            **{**(context.resource_config or {}), **(nested_resources or {})}
        )._initialize_and_run_cm(context) as value:
            yield value


class ConfigurableResource(ConfigurableResourceFactory[TResValue]):
    """Base class for Dagster resources that utilize structured config.

    This class is a subclass of both :py:class:`ResourceDefinition` and :py:class:`Config`.

    Example definition:

    .. code-block:: python

        class WriterResource(ConfigurableResource):
            prefix: str

            def output(self, text: str) -> None:
                print(f"{self.prefix}{text}")

    Example usage:

    .. code-block:: python

        @asset
        def asset_that_uses_writer(writer: WriterResource):
            writer.output("text")

        defs = Definitions(
            assets=[asset_that_uses_writer],
            resources={"writer": WriterResource(prefix="a_prefix")},
        )

    You can optionally use this class to model configuration only and vend an object
    of a different type for use at runtime. This is useful for those who wish to
    have a separate object that manages configuration and a separate object at runtime. Or
    where you want to directly use a third-party class that you do not control.

    To do this you override the `create_resource` methods to return a different object.

    .. code-block:: python

        class WriterResource(ConfigurableResource):
            str: prefix

            def create_resource(self, context: InitResourceContext) -> Writer:
                # Writer is pre-existing class defined else
                return Writer(self.prefix)

    Example usage:

    .. code-block:: python

        @asset
        def use_preexisting_writer_as_resource(writer: ResourceParam[Writer]):
            writer.output("text")

        defs = Definitions(
            assets=[use_preexisting_writer_as_resource],
            resources={"writer": WriterResource(prefix="a_prefix")},
        )
    """

    def create_resource(self, context: InitResourceContext) -> TResValue:
        """Returns the object that this resource hands to user code, accessible by ops or assets
        through the context or resource parameters. This works like the function decorated
        with @resource when using function-based resources.

        For ConfigurableResource, this function will return itself, passing
        the actual ConfigurableResource object to user code.
        """
        return cast(TResValue, self)


def _is_fully_configured(resource: "CoercibleToResource") -> bool:
    from dagster._core.execution.build_resources import wrap_resource_for_execution

    actual_resource = wrap_resource_for_execution(resource)
    res = (
        validate_config(
            actual_resource.config_schema.config_type,
            (
                actual_resource.config_schema.default_value
                if actual_resource.config_schema.default_provided
                else {}
            ),
        ).success
        is True
    )

    return res and not isinstance(resource, PartialResource)


class PartialResourceState(NamedTuple):
    nested_partial_resources: dict[str, Any]
    config_schema: DagsterField
    resource_fn: Callable[[InitResourceContext], Any]
    description: Optional[str]
    nested_resources: dict[str, Any]


class PartialResource(
    MakeConfigCacheable,
    Generic[TResValue],
):
    data: dict[str, Any]
    resource_cls: type[Any]

    def __init__(
        self,
        resource_cls: type[ConfigurableResourceFactory[TResValue]],
        data: dict[str, Any],
    ):
        resource_pointers, _data_without_resources = separate_resource_params(resource_cls, data)

        super().__init__(data=data, resource_cls=resource_cls)  # type: ignore  # extends BaseModel, takes kwargs

        def resource_fn(context: InitResourceContext):
            to_populate = resource_cls._get_non_default_public_field_values_cls(  # noqa: SLF001
                {**data, **context.resource_config}
            )
            instantiated = resource_cls(
                **to_populate
            )  # So that collisions are resolved in favor of the latest provided run config
            return instantiated._get_initialize_and_run_fn()(context)  # noqa: SLF001

        self._state__internal__ = PartialResourceState(
            # We keep track of any resources we depend on which are not fully configured
            # so that we can retrieve them at runtime
            nested_partial_resources={
                k: v for k, v in resource_pointers.items() if (not _is_fully_configured(v))
            },
            config_schema=infer_schema_from_config_class(
                resource_cls, fields_to_omit=set(resource_pointers.keys())
            ),
            resource_fn=resource_fn,
            description=resource_cls.__doc__,
            nested_resources={k: v for k, v in resource_pointers.items()},
        )

    # to make AllowDelayedDependencies work
    @property
    def _nested_partial_resources(
        self,
    ) -> Mapping[str, Any]:
        return self._state__internal__.nested_partial_resources

    @property
    def nested_resources(
        self,
    ) -> Mapping[str, Any]:
        return self._state__internal__.nested_resources

    @cached_method  # resource resolution depends on always resolving to the same ResourceDefinition instance
    def get_resource_definition(self) -> ConfigurableResourceFactoryResourceDefinition:
        return ConfigurableResourceFactoryResourceDefinition(
            self.resource_cls,
            resource_fn=self._state__internal__.resource_fn,
            config_schema=self._state__internal__.config_schema,
            description=self._state__internal__.description,
            nested_resources=self.nested_resources,
            nested_partial_resources=self._nested_partial_resources,
            dagster_maintained=self.resource_cls._is_dagster_maintained(),  # noqa: SLF001
        )


ResourceOrPartial: TypeAlias = Union[
    ConfigurableResourceFactory[TResValue], PartialResource[TResValue]
]
ResourceOrPartialOrValue: TypeAlias = Union[
    ConfigurableResourceFactory[TResValue],
    PartialResource[TResValue],
    ResourceDefinition,
    TResValue,
]


V = TypeVar("V")


class ResourceDependency(Generic[V]):
    def __set_name__(self, _owner, name):
        self._name = name

    def __get__(self, obj: "ConfigurableResourceFactory", owner: Any) -> V:
        return getattr(obj, self._name)

    def __set__(self, obj: Optional[object], value: ResourceOrPartialOrValue[V]) -> None:
        setattr(obj, self._name, value)


class ConfigurableLegacyResourceAdapter(ConfigurableResource, ABC):
    """Adapter base class for wrapping a decorated, function-style resource
    with structured config.

    To use this class, subclass it, define config schema fields using Pydantic,
    and implement the ``wrapped_resource`` method.

    Example:
    .. code-block:: python

        @resource(config_schema={"prefix": str})
        def writer_resource(context):
            prefix = context.resource_config["prefix"]

            def output(text: str) -> None:
                out_txt.append(f"{prefix}{text}")

            return output

        class WriterResource(ConfigurableLegacyResourceAdapter):
            prefix: str

            @property
            def wrapped_resource(self) -> ResourceDefinition:
                return writer_resource
    """

    @property
    @abstractmethod
    def wrapped_resource(self) -> ResourceDefinition:
        raise NotImplementedError()

    @cached_method
    def get_resource_definition(self) -> ConfigurableResourceFactoryResourceDefinition:
        return ConfigurableResourceFactoryResourceDefinition(
            self.__class__,
            resource_fn=self.wrapped_resource.resource_fn,
            config_schema=self._config_schema,
            description=self.__doc__,
            nested_resources=self.nested_resources,
            nested_partial_resources=self._nested_partial_resources,
            dagster_maintained=self._is_dagster_maintained(),
        )

    def __call__(self, *args, **kwargs):
        return self.wrapped_resource(*args, **kwargs)


class SeparatedResourceParams(NamedTuple):
    resources: dict[str, Any]
    non_resources: dict[str, Any]


def _is_annotated_as_resource_type(annotation: type, metadata: list[str]) -> bool:
    """Determines if a field in a structured config class is annotated as a resource type or not."""
    from dagster._config.pythonic_config.type_check_utils import safe_is_subclass

    if metadata and metadata[0] == "resource_dependency":
        return True

    if is_closed_python_optional_type(annotation):
        args = get_args(annotation)
        annotation_inner = next((arg for arg in args if arg is not None), None)
        if not annotation_inner:
            return False
        return _is_annotated_as_resource_type(annotation_inner, [])

    is_annotated_as_resource_dependency = get_origin(annotation) == ResourceDependency or getattr(
        annotation, "__metadata__", None
    ) == ("resource_dependency",)

    return is_annotated_as_resource_dependency or safe_is_subclass(
        annotation, (ResourceDefinition, ConfigurableResourceFactory)
    )


class ResourceDataWithAnnotation(NamedTuple):
    key: str
    value: Any
    annotation: type
    annotation_metadata: list[str]


def _get_resource_param_fields(cls: type[BaseModel]) -> set[str]:
    """Returns the set of field names in a structured config class which are annotated as resource types."""
    # We need to grab metadata from the annotation in order to tell if
    # this key was annotated with a typing.Annotated annotation (which we use for resource/resource deps),
    # since Pydantic 2.0 strips that info out and sticks any Annotated metadata in the
    # metadata field
    fields_by_resolved_field_name = {
        field.alias if field.alias else key: field for key, field in model_fields(cls).items()
    }

    return {
        field_name
        for field_name in fields_by_resolved_field_name
        if _is_annotated_as_resource_type(
            fields_by_resolved_field_name[field_name].annotation,
            fields_by_resolved_field_name[field_name].metadata,
        )
    }


def separate_resource_params(cls: type[BaseModel], data: dict[str, Any]) -> SeparatedResourceParams:
    """Separates out the key/value inputs of fields in a structured config Resource class which
    are marked as resources (ie, using ResourceDependency) from those which are not.
    """
    nested_resource_field_names = _get_resource_param_fields(cls)

    resources = {}
    non_resources = {}
    for field_name, field_value in data.items():
        if field_name in nested_resource_field_names:
            resources[field_name] = field_value
        else:
            non_resources[field_name] = field_value

    out = SeparatedResourceParams(
        resources=resources,
        non_resources=non_resources,
    )
    return out


def _call_resource_fn_with_default(
    stack: contextlib.ExitStack, obj: ResourceDefinition, context: InitResourceContext
) -> Any:
    from dagster._config.validate import process_config

    if isinstance(obj.config_schema, ConfiguredDefinitionConfigSchema):
        value = cast(dict[str, Any], obj.config_schema.resolve_config({}).value)
        context = context.replace_config(value["config"])
    elif obj.config_schema.default_provided:
        # To explain why we need to process config here;
        # - The resource available on the init context (context.resource_config) has already been processed
        # - The nested resource's config has also already been processed, but is only available in the broader run config dictionary.
        # - The only information we have access to here is the unprocessed default value, so we need to process it a second time.
        unprocessed_config = obj.config_schema.default_value
        evr = process_config(
            {"config": obj.config_schema.config_type}, {"config": unprocessed_config}
        )
        if not evr.success:
            raise DagsterInvalidConfigError(
                "Error in config for nested resource ",
                evr.errors,
                unprocessed_config,
            )
        context = context.replace_config(cast(dict, evr.value)["config"])

    if has_at_least_one_parameter(obj.resource_fn):
        result = cast(ResourceFunctionWithContext, obj.resource_fn)(context)
    else:
        result = cast(ResourceFunctionWithoutContext, obj.resource_fn)()

    is_fn_generator = (
        inspect.isgenerator(obj.resource_fn)
        or isinstance(obj.resource_fn, contextlib.ContextDecorator)
        or isinstance(result, contextlib.AbstractContextManager)
    )
    if is_fn_generator:
        return stack.enter_context(cast(contextlib.AbstractContextManager, result))
    else:
        return result


LateBoundTypesForResourceTypeChecking.set_actual_types_for_type_checking(
    resource_dep_type=ResourceDependency,
    resource_type=ConfigurableResourceFactory,
    partial_resource_type=PartialResource,
)


def validate_resource_annotated_function(fn) -> None:
    """Validates any parameters on the decorated function that are annotated with
    :py:class:`dagster.ResourceDefinition`, raising a :py:class:`dagster.DagsterInvalidDefinitionError`
    if any are not also instances of :py:class:`dagster.ConfigurableResource` (these resources should
    instead be wrapped in the :py:func:`dagster.Resource` Annotation).
    """
    from dagster import DagsterInvalidDefinitionError
    from dagster._config.pythonic_config.resource import (
        ConfigurableResource,
        ConfigurableResourceFactory,
    )
    from dagster._config.pythonic_config.type_check_utils import safe_is_subclass

    malformed_params = [
        param
        for param in get_function_params(fn)
        if safe_is_subclass(param.annotation, (ResourceDefinition, ConfigurableResourceFactory))
        and not safe_is_subclass(param.annotation, ConfigurableResource)
    ]
    if len(malformed_params) > 0:
        malformed_param = malformed_params[0]
        output_type = None
        if safe_is_subclass(malformed_param.annotation, ConfigurableResourceFactory):
            orig_bases = getattr(malformed_param.annotation, "__orig_bases__", ())
            for base in orig_bases:
                if get_origin(base) is ConfigurableResourceFactory:
                    args = get_args(base)
                    output_type = args[0] if len(args) == 1 else None
                    if output_type == TResValue:
                        output_type = None

        output_type_name = getattr(output_type, "__name__", str(output_type))
        raise DagsterInvalidDefinitionError(
            """Resource param '{param_name}' is annotated as '{annotation_type}', but '{annotation_type}' outputs {value_message} value to user code such as @ops and @assets. This annotation should instead be {annotation_suggestion}""".format(
                param_name=malformed_param.name,
                annotation_type=malformed_param.annotation,
                value_message=f"a '{output_type}'" if output_type else "an unknown",
                annotation_suggestion=(
                    f"'ResourceParam[{output_type_name}]'"
                    if output_type
                    else "'ResourceParam[Any]' or 'ResourceParam[<output type>]'"
                ),
            )
        )


CoercibleToResource: TypeAlias = Union[
    ResourceDefinition, ConfigurableResourceFactory, PartialResource
]


def is_coercible_to_resource(val: Any) -> TypeGuard[CoercibleToResource]:
    return isinstance(val, (ResourceDefinition, ConfigurableResourceFactory, PartialResource))


def coerce_to_resource(val: CoercibleToResource):
    if isinstance(val, ResourceDefinition):
        return val
    else:
        return val.get_resource_definition()


def _resolve_partial_resource_to_key(
    attr_name: str,
    partial_resource: CoercibleToResource,
    resource_defs: Mapping[str, ResourceDefinition],
) -> str:
    partial_def = coerce_to_resource(partial_resource)
    matches = [key for key, resource_def in resource_defs.items() if resource_def is partial_def]
    check.invariant(
        len(matches) == 1,
        f"Failed to find resource for {attr_name}, expected this to be caught when resource requirements where evaluated.",
    )
    return matches[0]


@record
class PartialResourceDependencyRequirement(ResourceRequirement):
    class_name: str
    attr_name: str
    partial_resource: CoercibleToResource

    def is_satisfied(self, resource_defs: Mapping[str, "ResourceDefinition"]):
        from dagster._config.pythonic_config.resource import coerce_to_resource

        return coerce_to_resource(self.partial_resource) in resource_defs.values()

    def ensure_satisfied(self, resource_defs: Mapping[str, "ResourceDefinition"]):
        if not self.is_satisfied(resource_defs):
            raise DagsterInvalidDefinitionError(
                f"Failed to resolve resource nested at {self.class_name}.{self.attr_name}. "
                "Any partially configured, nested resources must be provided as a top level resource."
            )



from collections.abc import Mapping
from functools import update_wrapper
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Optional,
    Union,
    overload,
)

import dagster._check as check
from dagster._core.decorator_utils import format_docstring_for_description
from dagster._core.definitions.config import ConfigMapping
from dagster._core.definitions.graph_definition import GraphDefinition
from dagster._core.definitions.hook_definition import HookDefinition
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.logger_definition import LoggerDefinition
from dagster._core.definitions.metadata import RawMetadataValue
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._utils.tags import normalize_tags

if TYPE_CHECKING:
    from dagster._core.definitions.executor_definition import ExecutorDefinition
    from dagster._core.definitions.partition import PartitionedConfig, PartitionsDefinition
    from dagster._core.definitions.run_config import RunConfig


class _Job:
    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Mapping[str, Any]] = None,
        run_tags: Optional[Mapping[str, Any]] = None,
        metadata: Optional[Mapping[str, RawMetadataValue]] = None,
        resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
        config: Optional[
            Union[ConfigMapping, Mapping[str, Any], "RunConfig", "PartitionedConfig"]
        ] = None,
        logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
        executor_def: Optional["ExecutorDefinition"] = None,
        hooks: Optional[AbstractSet[HookDefinition]] = None,
        op_retry_policy: Optional[RetryPolicy] = None,
        partitions_def: Optional["PartitionsDefinition"] = None,
        input_values: Optional[Mapping[str, object]] = None,
    ):
        from dagster._core.definitions.run_config import convert_config_input

        self.name = name
        self.description = description
        self.tags = normalize_tags(tags, warning_stacklevel=4)
        self.run_tags = run_tags
        self.metadata = metadata
        self.resource_defs = resource_defs
        self.config = convert_config_input(config)
        self.logger_defs = logger_defs
        self.executor_def = executor_def
        self.hooks = hooks
        self.op_retry_policy = op_retry_policy
        self.partitions_def = partitions_def
        self.input_values = input_values

    def __call__(self, fn: Callable[..., Any]) -> JobDefinition:
        check.callable_param(fn, "fn")

        if not self.name:
            self.name = fn.__name__

        from dagster._core.definitions.composition import do_composition

        (
            input_mappings,
            output_mappings,
            dependencies,
            node_defs,
            config_mapping,
            positional_inputs,
            input_assets,
        ) = do_composition(
            decorator_name="@job",
            graph_name=self.name,
            fn=fn,
            provided_input_defs=[],
            provided_output_defs=[],
            ignore_output_from_composition_fn=False,
            config_mapping=None,
        )

        graph_def = GraphDefinition(
            name=self.name,
            dependencies=dependencies,
            node_defs=node_defs,
            description=self.description or format_docstring_for_description(fn),
            input_mappings=input_mappings,
            output_mappings=output_mappings,
            config=config_mapping,
            positional_inputs=positional_inputs,
            tags=self.run_tags,
            input_assets=input_assets,
        )

        job_def = graph_def.to_job(
            description=self.description or format_docstring_for_description(fn),
            resource_defs=self.resource_defs,
            config=self.config,
            tags=self.tags,
            run_tags=self.run_tags,
            metadata=self.metadata,
            logger_defs=self.logger_defs,
            executor_def=self.executor_def,
            hooks=self.hooks,
            op_retry_policy=self.op_retry_policy,
            partitions_def=self.partitions_def,
            input_values=self.input_values,
        )
        update_wrapper(job_def, fn)
        return job_def


@overload
def job(compose_fn: Callable[..., Any]) -> JobDefinition: ...


@overload
def job(
    *,
    name: Optional[str] = ...,
    description: Optional[str] = ...,
    resource_defs: Optional[Mapping[str, object]] = ...,
    config: Union[ConfigMapping, Mapping[str, Any], "RunConfig", "PartitionedConfig"] = ...,
    tags: Optional[Mapping[str, Any]] = ...,
    run_tags: Optional[Mapping[str, Any]] = ...,
    metadata: Optional[Mapping[str, RawMetadataValue]] = ...,
    logger_defs: Optional[Mapping[str, LoggerDefinition]] = ...,
    executor_def: Optional["ExecutorDefinition"] = ...,
    hooks: Optional[AbstractSet[HookDefinition]] = ...,
    op_retry_policy: Optional[RetryPolicy] = ...,
    partitions_def: Optional["PartitionsDefinition"] = ...,
    input_values: Optional[Mapping[str, object]] = ...,
) -> _Job: ...


def job(
    compose_fn: Optional[Callable[..., Any]] = None,
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    resource_defs: Optional[Mapping[str, object]] = None,
    config: Optional[
        Union[ConfigMapping, Mapping[str, Any], "RunConfig", "PartitionedConfig"]
    ] = None,
    tags: Optional[Mapping[str, str]] = None,
    run_tags: Optional[Mapping[str, str]] = None,
    metadata: Optional[Mapping[str, RawMetadataValue]] = None,
    logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
    executor_def: Optional["ExecutorDefinition"] = None,
    hooks: Optional[AbstractSet[HookDefinition]] = None,
    op_retry_policy: Optional[RetryPolicy] = None,
    partitions_def: Optional["PartitionsDefinition"] = None,
    input_values: Optional[Mapping[str, object]] = None,
) -> Union[JobDefinition, _Job]:
    """Creates a job with the specified parameters from the decorated graph/op invocation function.

    Using this decorator allows you to build an executable job by writing a function that invokes
    ops (or graphs).

    Args:
        compose_fn (Callable[..., Any]:
            The decorated function. The body should contain op or graph invocations. Unlike op
            functions, does not accept a context argument.
        name (Optional[str]):
            The name for the Job. Defaults to the name of the this graph.
        resource_defs (Optional[Mapping[str, object]]):
            Resources that are required by this graph for execution.
            If not defined, `io_manager` will default to filesystem.
        config:
            Describes how the job is parameterized at runtime.

            If no value is provided, then the schema for the job's run config is a standard
            format based on its ops and resources.

            If a dictionary is provided, then it must conform to the standard config schema, and
            it will be used as the job's run config for the job whenever the job is executed.
            The values provided will be viewable and editable in the Dagster UI, so be
            careful with secrets.

            If a :py:class:`RunConfig` object is provided, then it will be used directly as the run config
            for the job whenever the job is executed, similar to providing a dictionary.

            If a :py:class:`ConfigMapping` object is provided, then the schema for the job's run config is
            determined by the config mapping, and the ConfigMapping, which should return
            configuration in the standard format to configure the job.

            If a :py:class:`PartitionedConfig` object is provided, then it defines a discrete set of config
            values that can parameterize the job, as well as a function for mapping those
            values to the base config. The values provided will be viewable and editable in the
            Dagster UI, so be careful with secrets.
        tags (Optional[Mapping[str, object]]): A set of key-value tags that annotate the job and can
            be used for searching and filtering in the UI. Values that are not already strings will
            be serialized as JSON. If `run_tags` is not set, then the content of `tags` will also be
            automatically appended to the tags of any runs of this job.
        run_tags (Optional[Mapping[str, object]]):
            A set of key-value tags that will be automatically attached to runs launched by this
            job. Values that are not already strings will be serialized as JSON. These tag values
            may be overwritten by tag values provided at invocation time. If `run_tags` is set, then
            `tags` are not automatically appended to the tags of any runs of this job.
        metadata (Optional[Dict[str, RawMetadataValue]]):
            Arbitrary information that will be attached to the JobDefinition and be viewable in the Dagster UI.
            Keys must be strings, and values must be python primitive types or one of the provided
            MetadataValue types
        logger_defs (Optional[Dict[str, LoggerDefinition]]):
            A dictionary of string logger identifiers to their implementations.
        executor_def (Optional[ExecutorDefinition]):
            How this Job will be executed. Defaults to :py:class:`multiprocess_executor` .
        op_retry_policy (Optional[RetryPolicy]): The default retry policy for all ops in this job.
            Only used if retry policy is not defined on the op definition or op invocation.
        partitions_def (Optional[PartitionsDefinition]): Defines a discrete set of partition keys
            that can parameterize the job. If this argument is supplied, the config argument
            can't also be supplied.
        input_values (Optional[Mapping[str, Any]]):
            A dictionary that maps python objects to the top-level inputs of a job.

    Examples:
        .. code-block:: python

            @op
            def return_one():
                return 1

            @op
            def add_one(in1):
                return in1 + 1

            @job
            def job1():
                add_one(return_one())
    """
    if compose_fn is not None:
        check.invariant(description is None)
        return _Job()(compose_fn)

    from dagster._core.execution.build_resources import wrap_resources_for_execution

    return _Job(
        name=name,
        description=description,
        resource_defs=wrap_resources_for_execution(resource_defs),
        config=config,
        tags=tags,
        run_tags=run_tags,
        metadata=metadata,
        logger_defs=logger_defs,
        executor_def=executor_def,
        hooks=hooks,
        op_retry_policy=op_retry_policy,
        partitions_def=partitions_def,
        input_values=input_values,
    )



import os
from typing import TYPE_CHECKING

"""This module contains the execution context objects that are internal to the system.
Not every property on these should be exposed to random Jane or Joe dagster user
so we have a different layer of objects that encode the explicit public API
in the user_context module.
"""

from collections.abc import Sequence
from dataclasses import dataclass
from hashlib import sha256
from typing import Optional

import dagster._check as check
from dagster._core.definitions.data_version import (
    DATA_VERSION_TAG,
    SKIP_PARTITION_DATA_VERSION_DEPENDENCY_THRESHOLD,
    extract_data_version_from_entry,
)
from dagster._core.definitions.events import AssetKey

if TYPE_CHECKING:
    from dagster._core.definitions.data_version import DataVersion
    from dagster._core.event_api import EventLogRecord
    from dagster._core.events import DagsterEventType
    from dagster._core.storage.event_log.base import AssetRecord


if TYPE_CHECKING:
    from dagster._core.execution.context.compute import StepExecutionContext


@dataclass
class InputAssetVersionInfo:
    # This is the storage id of the last materialization/observation of any partition of an asset.
    # Thus it is computed the same way for both partitioned and non-partitioned assets.
    storage_id: int

    # This is the event type of the event referenced by storage_id
    event_type: "DagsterEventType"

    # If the input asset is partitioned, this is a hash of the sorted data versions of each dependency
    # partition. If the input asset is not partitioned, this is the data version of the asset. It
    # can be none if we are sourcing a materialization from before data versions.
    data_version: Optional["DataVersion"]

    # This is the run_id on the event that the storage_id references
    run_id: str

    # This is the timestamp on the event that the storage_id references
    timestamp: float


class DataVersionCache:
    def __init__(self, context: "StepExecutionContext"):
        self._context = context
        self.input_asset_version_info: dict[AssetKey, Optional[InputAssetVersionInfo]] = {}
        self.is_external_input_asset_version_info_loaded = False
        self.values: dict[AssetKey, DataVersion] = {}

    def set_data_version(self, asset_key: AssetKey, data_version: "DataVersion") -> None:
        self.values[asset_key] = data_version

    def has_data_version(self, asset_key: AssetKey) -> bool:
        return asset_key in self.values

    def get_data_version(self, asset_key: AssetKey) -> "DataVersion":
        return self.values[asset_key]

    def maybe_fetch_and_get_input_asset_version_info(
        self, key: AssetKey
    ) -> Optional["InputAssetVersionInfo"]:
        if key not in self.input_asset_version_info:
            self._fetch_input_asset_version_info([key])
        return self.input_asset_version_info[key]

    # "external" refers to records for inputs generated outside of this step
    def fetch_external_input_asset_version_info(self) -> None:
        output_keys = self._context.get_output_asset_keys()

        all_dep_keys: list[AssetKey] = []
        for output_key in output_keys:
            if not self._context.job_def.asset_layer.has(output_key):
                continue
            dep_keys = self._context.job_def.asset_layer.get(output_key).parent_keys
            for key in dep_keys:
                if key not in all_dep_keys and key not in output_keys:
                    all_dep_keys.append(key)

        self.input_asset_version_info = {}
        self._fetch_input_asset_version_info(all_dep_keys)
        self.is_external_input_asset_version_info_loaded = True

    def _fetch_input_asset_version_info(self, asset_keys: Sequence[AssetKey]) -> None:
        from dagster._core.definitions.data_version import extract_data_version_from_entry

        asset_records_by_key = self._fetch_asset_records(asset_keys)
        for key in asset_keys:
            asset_record = asset_records_by_key.get(key)
            event = self._get_input_asset_event(key, asset_record)
            if event is None:
                self.input_asset_version_info[key] = None
            else:
                storage_id = event.storage_id
                # Input name will be none if this is an internal dep
                input_name = self._context.job_def.asset_layer.input_for_asset_key(
                    self._context.node_handle, key
                )
                # Exclude AllPartitionMapping for now to avoid huge queries
                if input_name and self._context.has_asset_partitions_for_input(input_name):
                    subset = self._context.asset_partitions_subset_for_input(
                        input_name, require_valid_partitions=False
                    )
                    input_keys = list(subset.get_partition_keys())

                    # This check represents a temporary constraint that prevents huge query results for upstream
                    # partition data versions from timing out runs. If a partitioned dependency (a) uses an
                    # AllPartitionMapping; and (b) has greater than or equal to
                    # SKIP_PARTITION_DATA_VERSION_DEPENDENCY_THRESHOLD dependency partitions, then we
                    # process it as a non-partitioned dependency (note that this was the behavior for
                    # all partition dependencies prior to 2023-08).  This means that stale status
                    # results cannot be accurately computed for the dependency, and there is thus
                    # corresponding logic in the CachingStaleStatusResolver to account for this. This
                    # constraint should be removed when we have thoroughly examined the performance of
                    # the data version retrieval query and can guarantee decent performance.
                    if len(input_keys) < SKIP_PARTITION_DATA_VERSION_DEPENDENCY_THRESHOLD:
                        data_version = self._get_partitions_data_version_from_keys(key, input_keys)
                    else:
                        data_version = extract_data_version_from_entry(event.event_log_entry)
                else:
                    data_version = extract_data_version_from_entry(event.event_log_entry)
                self.input_asset_version_info[key] = InputAssetVersionInfo(
                    storage_id,
                    check.not_none(event.event_log_entry.dagster_event).event_type,
                    data_version,
                    event.run_id,
                    event.timestamp,
                )

    def _fetch_asset_records(self, asset_keys: Sequence[AssetKey]) -> dict[AssetKey, "AssetRecord"]:
        batch_size = int(os.getenv("GET_ASSET_RECORDS_FOR_DATA_VERSION_BATCH_SIZE", "100"))
        asset_records_by_key = {}
        to_fetch = asset_keys
        while len(to_fetch):
            for record in self._context.instance.get_asset_records(to_fetch[:batch_size]):
                asset_records_by_key[record.asset_entry.asset_key] = record
            to_fetch = to_fetch[batch_size:]

        return asset_records_by_key

    def _get_input_asset_event(
        self, key: AssetKey, asset_record: Optional["AssetRecord"]
    ) -> Optional["EventLogRecord"]:
        event = None
        if asset_record and asset_record.asset_entry.last_materialization_record:
            event = asset_record.asset_entry.last_materialization_record
        elif (
            asset_record
            and self._context.instance.event_log_storage.asset_records_have_last_observation
        ):
            event = asset_record.asset_entry.last_observation_record

        if (
            not event
            and not self._context.instance.event_log_storage.asset_records_have_last_observation
        ):
            event = next(
                iter(self._context.instance.fetch_observations(key, limit=1).records), None
            )

        if event:
            self._check_input_asset_event(key, event)
        return event

    def _check_input_asset_event(self, key: AssetKey, event: "EventLogRecord") -> None:
        assert event.event_log_entry
        event_data_version = extract_data_version_from_entry(event.event_log_entry)
        if key in self.values and self.values[key] != event_data_version:
            self._context.log.warning(
                f"Data version mismatch for asset {key}. Data version from materialization within"
                f" current step is `{self.values[key]}`. Data version from most recent"
                f" materialization is `{event_data_version}`. Most recent materialization will be"
                " used for provenance tracking."
            )

    def _get_partitions_data_version_from_keys(
        self, key: AssetKey, partition_keys: Sequence[str]
    ) -> "DataVersion":
        from dagster._core.definitions.data_version import DataVersion
        from dagster._core.events import DagsterEventType

        # TODO: this needs to account for observations also
        event_type = DagsterEventType.ASSET_MATERIALIZATION
        tags_by_partition = self._context.instance._event_storage.get_latest_tags_by_partition(  # noqa: SLF001
            key, event_type, [DATA_VERSION_TAG], asset_partitions=list(partition_keys)
        )
        partition_data_versions = [
            pair[1][DATA_VERSION_TAG]
            for pair in sorted(tags_by_partition.items(), key=lambda x: x[0])
        ]
        hash_sig = sha256()
        hash_sig.update(bytearray("".join(partition_data_versions), "utf8"))
        return DataVersion(hash_sig.hexdigest())

    # Call this to clear the cache for an input asset record. This is necessary when an old
    # materialization for an asset was loaded during `fetch_external_input_asset_records` because an
    # intrastep asset is not required, but then that asset is materialized during the step. If we
    # don't clear the cache for this asset, then we won't use the most up-to-date asset record.
    def wipe_input_asset_version_info(self, key: AssetKey) -> None:
        if key in self.input_asset_version_info:
            del self.input_asset_version_info[key]



from collections.abc import Iterable, Sequence
from functools import total_ordering
from heapq import heapify, heappop, heappush
from typing import TYPE_CHECKING, Callable, NamedTuple

import dagster._check as check
from dagster._core.asset_graph_view.asset_graph_view import AssetGraphView
from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.asset_graph_subset import AssetGraphSubset

if TYPE_CHECKING:
    from dagster._core.asset_graph_view.entity_subset import EntitySubset


class AssetGraphViewBfsFilterConditionResult(NamedTuple):
    passed_asset_graph_subset: AssetGraphSubset
    excluded_asset_graph_subsets_and_reasons: Sequence[tuple[AssetGraphSubset, str]]


def bfs_filter_asset_graph_view(
    asset_graph_view: AssetGraphView,
    condition_fn: Callable[
        ["AssetGraphSubset", "AssetGraphSubset"],
        AssetGraphViewBfsFilterConditionResult,
    ],
    initial_asset_graph_subset: "AssetGraphSubset",
    include_full_execution_set: bool,
) -> tuple[AssetGraphSubset, Sequence[tuple[AssetGraphSubset, str]]]:
    """Returns the subset of the graph that satisfy supplied criteria.

    - Are >= initial_asset_graph_subset
    - Match the condition_fn
    - Any of their ancestors >= initial_asset_graph_subset match the condition_fn

    Also returns a list of tuples, where each tuple is an asset subset that did not
    satisfy the condition and the reason they were filtered out.

    The condition_fn takes in:
    - a subset of the asset graph to evaluate the condition for. If include_full_execution_set=True,
    the asset keys are all part of the same execution set (i.e. non-subsettable multi-asset). If
    include_full_execution_set=False, only a single asset key will be in the subset.

    - An AssetGraphSubset for the portion of the graph that has so far been visited and passed
    the condition.

    The condition_fn should return a object with an AssetGraphSubset indicating the portion
    of the subset that passes the condition, and a list of (AssetGraphSubset, str)
    tuples with more information about why certain subsets were excluded.

    Visits parents before children.
    """
    initial_subsets = list(asset_graph_view.iterate_asset_subsets(initial_asset_graph_subset))

    # invariant: we never consider an asset partition before considering its ancestors
    queue = ToposortedPriorityQueue(
        asset_graph_view, initial_subsets, include_full_execution_set=include_full_execution_set
    )

    visited_graph_subset = initial_asset_graph_subset

    result: AssetGraphSubset = AssetGraphSubset.create_empty_subset()
    failed_reasons: Sequence[tuple[AssetGraphSubset, str]] = []

    asset_graph = asset_graph_view.asset_graph

    while len(queue) > 0:
        candidate_subset = queue.dequeue()
        condition_result = condition_fn(candidate_subset, result)

        subset_that_meets_condition = condition_result.passed_asset_graph_subset
        failed_reasons.extend(condition_result.excluded_asset_graph_subsets_and_reasons)

        result = result | subset_that_meets_condition

        for matching_entity_subset in asset_graph_view.iterate_asset_subsets(
            subset_that_meets_condition
        ):
            # Add any child subsets that have not yet been visited to the queue
            for child_key in asset_graph.get(matching_entity_subset.key).child_keys:
                child_subset = asset_graph_view.compute_child_subset(
                    child_key, matching_entity_subset
                )
                unvisited_child_subset = child_subset.compute_difference(
                    asset_graph_view.get_entity_subset_from_asset_graph_subset(
                        visited_graph_subset, child_key
                    )
                )
                if not unvisited_child_subset.is_empty:
                    queue.enqueue(unvisited_child_subset)
                    visited_graph_subset = (
                        visited_graph_subset
                        | AssetGraphSubset.from_entity_subsets([unvisited_child_subset])
                    )

    return result, failed_reasons


class ToposortedPriorityQueue:
    """Queue that returns parents before their children."""

    @total_ordering
    class QueueItem(NamedTuple):
        level: int
        sort_key: str
        asset_graph_subset: AssetGraphSubset

        def __eq__(self, other: object) -> bool:
            if isinstance(other, ToposortedPriorityQueue.QueueItem):
                return self.level == other.level and self.sort_key == other.sort_key
            return False

        def __lt__(self, other: object) -> bool:
            if isinstance(other, ToposortedPriorityQueue.QueueItem):
                return self.level < other.level or (
                    self.level == other.level and self.sort_key < other.sort_key
                )
            raise TypeError()

    def __init__(
        self,
        asset_graph_view: AssetGraphView,
        items: Iterable["EntitySubset"],
        include_full_execution_set: bool,
    ):
        self._asset_graph_view = asset_graph_view
        self._include_full_execution_set = include_full_execution_set

        self._toposort_level_by_asset_key = {
            asset_key: i
            for i, asset_keys in enumerate(
                asset_graph_view.asset_graph.toposorted_asset_keys_by_level
            )
            for asset_key in asset_keys
        }
        self._heap = [self._queue_item(entity_subset) for entity_subset in items]
        heapify(self._heap)

    def enqueue(self, entity_subset: "EntitySubset") -> None:
        heappush(self._heap, self._queue_item(entity_subset))

    def dequeue(self) -> AssetGraphSubset:
        # For multi-assets, will include all required multi-asset keys if
        # include_full_execution_set is set to True, or just the passed in
        # asset key if it was not. If there are multiple assets in the subset
        # the subset will have the same partitions included for each asset.
        heap_value = heappop(self._heap)
        return heap_value.asset_graph_subset

    def _queue_item(self, entity_subset: "EntitySubset") -> "ToposortedPriorityQueue.QueueItem":
        asset_key = entity_subset.key

        if self._include_full_execution_set:
            execution_set_keys = self._asset_graph_view.asset_graph.get(
                asset_key
            ).execution_set_asset_keys
        else:
            execution_set_keys = {asset_key}

        level = max(
            self._toposort_level_by_asset_key[asset_key] for asset_key in execution_set_keys
        )

        serializable_entity_subset = entity_subset.convert_to_serializable_subset()

        serializable_entity_subsets = [
            SerializableEntitySubset(key=asset_key, value=serializable_entity_subset.value)
            for asset_key in execution_set_keys
        ]

        entity_subsets = [
            check.not_none(
                self._asset_graph_view.get_subset_from_serializable_subset(
                    serializable_entity_subset
                )
            )
            for serializable_entity_subset in serializable_entity_subsets
        ]

        asset_graph_subset = AssetGraphSubset.from_entity_subsets(entity_subsets)

        return ToposortedPriorityQueue.QueueItem(
            level,
            asset_key.to_string(),
            asset_graph_subset=asset_graph_subset,
        )

    def __len__(self) -> int:
        return len(self._heap)



from collections.abc import Iterator, Mapping, Sequence
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, NamedTuple, Optional, Union

from dagster import _check as check
from dagster._core.definitions import AssetKey
from dagster._core.definitions.asset_graph_subset import AssetGraphSubset
from dagster._core.definitions.base_asset_graph import BaseAssetGraph
from dagster._core.definitions.partition import PartitionsSubset
from dagster._core.definitions.run_request import RunRequest
from dagster._core.definitions.selector import PartitionsByAssetSelector
from dagster._core.definitions.utils import check_valid_title
from dagster._core.errors import DagsterDefinitionChangedDeserializationError
from dagster._core.execution.asset_backfill import (
    AssetBackfillData,
    PartitionedAssetBackfillStatus,
    UnpartitionedAssetBackfillStatus,
)
from dagster._core.execution.bulk_actions import BulkActionType
from dagster._core.instance import DynamicPartitionsStore
from dagster._core.remote_representation.external_data import job_name_for_partition_set_snap_name
from dagster._core.remote_representation.origin import RemotePartitionSetOrigin
from dagster._core.storage.dagster_run import (
    CANCELABLE_RUN_STATUSES,
    NOT_FINISHED_STATUSES,
    RunsFilter,
)
from dagster._core.storage.tags import BACKFILL_ID_TAG, USER_TAG
from dagster._core.workspace.context import BaseWorkspaceRequestContext
from dagster._record import record
from dagster._serdes import whitelist_for_serdes
from dagster._utils.error import SerializableErrorInfo

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance

MAX_RUNS_CANCELED_PER_ITERATION = 50


@whitelist_for_serdes
class BulkActionStatus(Enum):
    REQUESTED = "REQUESTED"
    COMPLETED = "COMPLETED"  # deprecated. Use COMPLETED_SUCCESS or COMPLETED_FAILED instead
    FAILED = "FAILED"  # denotes when there is a daemon failure, or some other issue processing the backfill
    CANCELING = "CANCELING"
    CANCELED = "CANCELED"
    COMPLETED_SUCCESS = "COMPLETED_SUCCESS"
    COMPLETED_FAILED = "COMPLETED_FAILED"  # denotes that the backfill daemon completed successfully, but some runs failed

    @staticmethod
    def from_graphql_input(graphql_str):
        return BulkActionStatus(graphql_str)


BULK_ACTION_TERMINAL_STATUSES = [
    BulkActionStatus.COMPLETED,
    BulkActionStatus.FAILED,
    BulkActionStatus.CANCELED,
    BulkActionStatus.COMPLETED_SUCCESS,
    BulkActionStatus.COMPLETED_SUCCESS,
    BulkActionStatus.COMPLETED_FAILED,
]


@record
class BulkActionsFilter:
    """Filters to use when querying for bulk actions (i.e. backfills) from the BulkActionsTable.

    Each field of the BulkActionsFilter represents a logical AND with each other. For
    example, if you specify status and created_before, then you will receive only bulk actions
    with the specified states AND the created before created_before. If left blank, then
    all values will be permitted for that field.

    Args:
        statuses (Optional[Sequence[BulkActionStatus]]): A list of statuses to filter by.
        created_before (Optional[DateTime]): Filter by bulk actions that were created before this datetime. Note that the
            create_time for each bulk action is stored in UTC.
        created_after (Optional[DateTime]): Filter by bulk actions that were created after this datetime. Note that the
            create_time for each bulk action is stored in UTC.
        tags (Optional[Dict[str, Union[str, List[str]]]]): A dictionary of tags to query by. All tags specified
            here must be present for a given bulk action to pass the filter.
        job_name (Optional[str]): Name of the job to query for. If blank, all job_names will be accepted.
        backfill_ids (Optional[Sequence[str]]): A list of backfill_ids to filter by. If blank, all backfill_ids will be included
    """

    statuses: Optional[Sequence[BulkActionStatus]] = None
    created_before: Optional[datetime] = None
    created_after: Optional[datetime] = None
    tags: Optional[Mapping[str, Union[str, Sequence[str]]]] = None
    job_name: Optional[str] = None
    backfill_ids: Optional[Sequence[str]] = None


@whitelist_for_serdes
class PartitionBackfill(
    NamedTuple(
        "_PartitionBackfill",
        [
            ("backfill_id", str),
            ("status", BulkActionStatus),
            ("from_failure", bool),
            ("tags", Mapping[str, str]),
            ("backfill_timestamp", float),
            ("error", Optional[SerializableErrorInfo]),
            ("asset_selection", Optional[Sequence[AssetKey]]),
            ("title", Optional[str]),
            ("description", Optional[str]),
            # fields that are only used by job backfills
            ("partition_set_origin", Optional[RemotePartitionSetOrigin]),
            ("partition_names", Optional[Sequence[str]]),
            ("last_submitted_partition_name", Optional[str]),
            ("reexecution_steps", Optional[Sequence[str]]),
            # only used by asset backfills
            ("serialized_asset_backfill_data", Optional[str]),
            ("asset_backfill_data", Optional[AssetBackfillData]),
            ("failure_count", int),
            ("submitting_run_requests", Sequence[RunRequest]),
            ("reserved_run_ids", Sequence[str]),
            ("backfill_end_timestamp", Optional[float]),
        ],
    ),
):
    def __new__(
        cls,
        backfill_id: str,
        status: BulkActionStatus,
        from_failure: bool,
        tags: Optional[Mapping[str, str]],
        backfill_timestamp: float,
        error: Optional[SerializableErrorInfo] = None,
        asset_selection: Optional[Sequence[AssetKey]] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        partition_set_origin: Optional[RemotePartitionSetOrigin] = None,
        partition_names: Optional[Sequence[str]] = None,
        last_submitted_partition_name: Optional[str] = None,
        reexecution_steps: Optional[Sequence[str]] = None,
        serialized_asset_backfill_data: Optional[str] = None,
        asset_backfill_data: Optional[AssetBackfillData] = None,
        failure_count: Optional[int] = None,
        submitting_run_requests: Optional[Sequence[RunRequest]] = None,
        reserved_run_ids: Optional[Sequence[str]] = None,
        backfill_end_timestamp: Optional[float] = None,
    ):
        check.invariant(
            not (asset_selection and reexecution_steps),
            "Can't supply both an asset_selection and reexecution_steps to a PartitionBackfill.",
        )

        if serialized_asset_backfill_data is not None:
            check.invariant(partition_set_origin is None)
            check.invariant(partition_names is None)
            check.invariant(last_submitted_partition_name is None)
            check.invariant(reexecution_steps is None)

        return super().__new__(
            cls,
            backfill_id=check.str_param(backfill_id, "backfill_id"),
            status=check.inst_param(status, "status", BulkActionStatus),
            from_failure=check.bool_param(from_failure, "from_failure"),
            tags=check.opt_mapping_param(tags, "tags", key_type=str, value_type=str),
            backfill_timestamp=check.float_param(backfill_timestamp, "backfill_timestamp"),
            error=check.opt_inst_param(error, "error", SerializableErrorInfo),
            asset_selection=check.opt_nullable_sequence_param(
                asset_selection, "asset_selection", of_type=AssetKey
            ),
            title=check_valid_title(title),
            description=check.opt_str_param(description, "description"),
            partition_set_origin=check.opt_inst_param(
                partition_set_origin, "partition_set_origin", RemotePartitionSetOrigin
            ),
            partition_names=check.opt_nullable_sequence_param(
                partition_names, "partition_names", of_type=str
            ),
            last_submitted_partition_name=check.opt_str_param(
                last_submitted_partition_name, "last_submitted_partition_name"
            ),
            reexecution_steps=check.opt_nullable_sequence_param(
                reexecution_steps, "reexecution_steps", of_type=str
            ),
            serialized_asset_backfill_data=check.opt_str_param(
                serialized_asset_backfill_data, "serialized_asset_backfill_data"
            ),
            asset_backfill_data=check.opt_inst_param(
                asset_backfill_data, "asset_backfill_data", AssetBackfillData
            ),
            failure_count=check.opt_int_param(failure_count, "failure_count", 0),
            submitting_run_requests=check.opt_sequence_param(
                submitting_run_requests, "submitting_run_requests", of_type=RunRequest
            ),
            reserved_run_ids=check.opt_sequence_param(
                reserved_run_ids, "reserved_run_ids", of_type=str
            ),
            backfill_end_timestamp=check.opt_float_param(
                backfill_end_timestamp, "backfill_end_timestamp"
            ),
        )

    @property
    def selector_id(self):
        return self.partition_set_origin.get_selector_id() if self.partition_set_origin else None

    @property
    def is_asset_backfill(self) -> bool:
        return (
            self.serialized_asset_backfill_data is not None or self.asset_backfill_data is not None
        )

    def get_asset_backfill_data(self, asset_graph: BaseAssetGraph) -> AssetBackfillData:
        if self.serialized_asset_backfill_data:
            asset_backfill_data = AssetBackfillData.from_serialized(
                self.serialized_asset_backfill_data, asset_graph, self.backfill_timestamp
            )
        elif self.asset_backfill_data:
            asset_backfill_data = self.asset_backfill_data
        else:
            check.failed("Expected either serialized_asset_backfill_data or asset_backfill_data")

        return asset_backfill_data

    @property
    def bulk_action_type(self) -> BulkActionType:
        if self.is_asset_backfill:
            return BulkActionType.MULTI_RUN_ASSET_ACTION
        else:
            return BulkActionType.PARTITION_BACKFILL

    @property
    def partition_set_name(self) -> Optional[str]:
        if self.partition_set_origin is None:
            return None

        return self.partition_set_origin.partition_set_name

    @property
    def job_name(self) -> Optional[str]:
        if self.is_asset_backfill:
            return None
        return (
            job_name_for_partition_set_snap_name(self.partition_set_name)
            if self.partition_set_name
            else None
        )

    @property
    def log_storage_prefix(self) -> Sequence[str]:
        return ["backfill", self.backfill_id]

    @property
    def user(self) -> Optional[str]:
        if self.tags:
            return self.tags.get(USER_TAG)
        return None

    def is_valid_serialization(self, workspace: BaseWorkspaceRequestContext) -> bool:
        if self.is_asset_backfill:
            if self.serialized_asset_backfill_data:
                return AssetBackfillData.is_valid_serialization(
                    self.serialized_asset_backfill_data,
                    workspace.asset_graph,
                )
            else:
                return True
        else:
            return True

    def get_backfill_status_per_asset_key(
        self, workspace: BaseWorkspaceRequestContext
    ) -> Sequence[Union[PartitionedAssetBackfillStatus, UnpartitionedAssetBackfillStatus]]:
        """Returns a sequence of backfill statuses for each targeted asset key in the asset graph,
        in topological order.
        """
        if not self.is_valid_serialization(workspace):
            return []

        if self.is_asset_backfill:
            asset_graph = workspace.asset_graph
            try:
                asset_backfill_data = self.get_asset_backfill_data(asset_graph)
            except DagsterDefinitionChangedDeserializationError:
                return []

            return asset_backfill_data.get_backfill_status_per_asset_key(asset_graph)
        else:
            return []

    def get_target_partitions_subset(
        self, workspace: BaseWorkspaceRequestContext, asset_key: AssetKey
    ) -> Optional[PartitionsSubset]:
        if not self.is_valid_serialization(workspace):
            return None

        if self.is_asset_backfill:
            asset_graph = workspace.asset_graph
            try:
                asset_backfill_data = self.get_asset_backfill_data(asset_graph)
            except DagsterDefinitionChangedDeserializationError:
                return None

            return asset_backfill_data.get_target_partitions_subset(asset_key)
        else:
            return None

    def get_target_root_partitions_subset(
        self, workspace: BaseWorkspaceRequestContext
    ) -> Optional[PartitionsSubset]:
        if not self.is_valid_serialization(workspace):
            return None

        if self.is_asset_backfill:
            asset_graph = workspace.asset_graph
            try:
                asset_backfill_data = self.get_asset_backfill_data(asset_graph)
            except DagsterDefinitionChangedDeserializationError:
                return None

            return asset_backfill_data.get_target_root_partitions_subset(asset_graph)
        else:
            return None

    def get_num_partitions(self, workspace: BaseWorkspaceRequestContext) -> Optional[int]:
        if not self.is_valid_serialization(workspace):
            return 0

        if self.is_asset_backfill:
            asset_graph = workspace.asset_graph
            try:
                asset_backfill_data = self.get_asset_backfill_data(asset_graph)
            except DagsterDefinitionChangedDeserializationError:
                return 0

            return asset_backfill_data.get_num_partitions()
        else:
            if self.partition_names is None:
                check.failed("Non-asset backfills should have a non-null partition_names field")

            return len(self.partition_names)

    def get_partition_names(
        self, workspace: BaseWorkspaceRequestContext
    ) -> Optional[Sequence[str]]:
        if not self.is_valid_serialization(workspace):
            return []

        if self.is_asset_backfill:
            asset_graph = workspace.asset_graph
            try:
                asset_backfill_data = self.get_asset_backfill_data(asset_graph)
            except DagsterDefinitionChangedDeserializationError:
                return None

            return asset_backfill_data.get_partition_names()
        else:
            if self.partition_names is None:
                check.failed("Non-asset backfills should have a non-null partition_names field")

            return self.partition_names

    def get_num_cancelable(self) -> int:
        """This method is only valid for job backfills. It eturns the number of partitions that are have
        not yet been requested by the backfill.

        For asset backfills, returns 0.
        """
        if self.is_asset_backfill:
            return 0

        if self.status != BulkActionStatus.REQUESTED:
            return 0

        if self.partition_names is None:
            check.failed("Expected partition_names to not be None for job backfill")

        checkpoint = self.last_submitted_partition_name
        total_count = len(self.partition_names)
        checkpoint_idx = (
            self.partition_names.index(checkpoint) + 1
            if checkpoint and checkpoint in self.partition_names
            else 0
        )
        return max(0, total_count - checkpoint_idx)

    def with_status(self, status):
        check.inst_param(status, "status", BulkActionStatus)
        return self._replace(status=status)

    def with_partition_checkpoint(self, last_submitted_partition_name):
        check.opt_str_param(last_submitted_partition_name, "last_submitted_partition_name")
        return self._replace(last_submitted_partition_name=last_submitted_partition_name)

    def with_submitting_run_requests(
        self, submitting_run_requests: Sequence[RunRequest], reserved_run_ids: Sequence[str]
    ) -> "PartitionBackfill":
        return self._replace(
            submitting_run_requests=submitting_run_requests,
            reserved_run_ids=reserved_run_ids,
        )

    def with_failure_count(self, new_failure_count: int):
        return self._replace(failure_count=new_failure_count)

    def with_error(self, error):
        check.opt_inst_param(error, "error", SerializableErrorInfo)
        return self._replace(error=error)

    def with_end_timestamp(self, end_timestamp: float) -> "PartitionBackfill":
        check.float_param(end_timestamp, "end_timestamp")
        return self._replace(backfill_end_timestamp=end_timestamp)

    def with_asset_backfill_data(
        self,
        asset_backfill_data: AssetBackfillData,
        dynamic_partitions_store: DynamicPartitionsStore,
        asset_graph: BaseAssetGraph,
    ) -> "PartitionBackfill":
        is_backcompat = self.serialized_asset_backfill_data is not None
        return self._replace(
            serialized_asset_backfill_data=(
                asset_backfill_data.serialize(
                    dynamic_partitions_store=dynamic_partitions_store, asset_graph=asset_graph
                )
                if is_backcompat
                else None
            ),
            asset_backfill_data=(asset_backfill_data if not is_backcompat else None),
        )

    @classmethod
    def from_asset_partitions(
        cls,
        backfill_id: str,
        asset_graph: BaseAssetGraph,
        partition_names: Optional[Sequence[str]],
        asset_selection: Sequence[AssetKey],
        backfill_timestamp: float,
        tags: Mapping[str, str],
        dynamic_partitions_store: DynamicPartitionsStore,
        all_partitions: bool,
        title: Optional[str],
        description: Optional[str],
    ) -> "PartitionBackfill":
        """If all the selected assets that have PartitionsDefinitions have the same partitioning, then
        the backfill will target the provided partition_names for all those assets.

        Otherwise, the backfill must consist of a partitioned "anchor" asset and a set of other
        assets that descend from it. In that case, the backfill will target the partition_names of
        the anchor asset, as well as all partitions of other selected assets that are downstream
        of those partitions of the anchor asset.
        """
        asset_backfill_data = AssetBackfillData.from_asset_partitions(
            asset_graph=asset_graph,
            partition_names=partition_names,
            asset_selection=asset_selection,
            dynamic_partitions_store=dynamic_partitions_store,
            all_partitions=all_partitions,
            backfill_start_timestamp=backfill_timestamp,
        )
        return cls(
            backfill_id=backfill_id,
            status=BulkActionStatus.REQUESTED,
            from_failure=False,
            tags=tags,
            backfill_timestamp=backfill_timestamp,
            asset_selection=asset_selection,
            serialized_asset_backfill_data=None,
            asset_backfill_data=asset_backfill_data,
            title=title,
            description=description,
        )

    @classmethod
    def from_partitions_by_assets(
        cls,
        backfill_id: str,
        asset_graph: BaseAssetGraph,
        backfill_timestamp: float,
        tags: Mapping[str, str],
        dynamic_partitions_store: DynamicPartitionsStore,
        partitions_by_assets: Sequence[PartitionsByAssetSelector],
        title: Optional[str],
        description: Optional[str],
    ):
        asset_backfill_data = AssetBackfillData.from_partitions_by_assets(
            asset_graph=asset_graph,
            dynamic_partitions_store=dynamic_partitions_store,
            backfill_start_timestamp=backfill_timestamp,
            partitions_by_assets=partitions_by_assets,
        )
        return cls(
            backfill_id=backfill_id,
            status=BulkActionStatus.REQUESTED,
            from_failure=False,
            tags=tags,
            backfill_timestamp=backfill_timestamp,
            serialized_asset_backfill_data=None,
            asset_backfill_data=asset_backfill_data,
            asset_selection=[selector.asset_key for selector in partitions_by_assets],
            title=title,
            description=description,
        )

    @classmethod
    def from_asset_graph_subset(
        cls,
        backfill_id: str,
        backfill_timestamp: float,
        tags: Mapping[str, str],
        dynamic_partitions_store: DynamicPartitionsStore,
        asset_graph_subset: AssetGraphSubset,
        title: Optional[str],
        description: Optional[str],
    ):
        asset_backfill_data = AssetBackfillData.from_asset_graph_subset(
            asset_graph_subset=asset_graph_subset,
            dynamic_partitions_store=dynamic_partitions_store,
            backfill_start_timestamp=backfill_timestamp,
        )
        return cls(
            backfill_id=backfill_id,
            status=BulkActionStatus.REQUESTED,
            from_failure=False,
            tags=tags,
            backfill_timestamp=backfill_timestamp,
            serialized_asset_backfill_data=None,
            asset_backfill_data=asset_backfill_data,
            asset_selection=list(asset_graph_subset.asset_keys),
            title=title,
            description=description,
        )


def cancel_backfill_runs_and_cancellation_complete(
    instance: "DagsterInstance", backfill_id: str
) -> Iterator[Union[None, bool]]:
    """Cancels MAX_RUNS_CANCELED_PER_ITERATION runs associated with the backfill_id. Ensures that
    all runs for the backfill are in a terminal state before indicating that the backfill can be marked
    CANCELED.
    Yields a boolean indicating the backfill can be considered canceled (ie all runs are canceled).
    """
    if not instance.run_coordinator:
        check.failed("The instance must have a run coordinator in order to cancel runs")

    # Query for cancelable runs, enforcing a limit on the number of runs to cancel in an iteration
    # as canceling runs incurs cost
    runs_to_cancel_in_iteration = instance.run_storage.get_run_ids(
        filters=RunsFilter(
            statuses=CANCELABLE_RUN_STATUSES,
            tags={
                BACKFILL_ID_TAG: backfill_id,
            },
        ),
        limit=MAX_RUNS_CANCELED_PER_ITERATION,
    )

    yield None

    if runs_to_cancel_in_iteration:
        # since we are canceling some runs in this iteration, we know that there is more work to do.
        # Either cancelling more runs, or waiting for the canceled runs to get to a terminal state
        work_done = False
        for run_id in runs_to_cancel_in_iteration:
            instance.run_coordinator.cancel_run(run_id)
            yield None
    else:
        # If there are no runs to cancel, check if there are any runs still in progress. If there are,
        # then we want to wait for them to reach a terminal state before the backfill is marked CANCELED.
        run_waiting_to_cancel = instance.get_run_ids(
            RunsFilter(
                tags={BACKFILL_ID_TAG: backfill_id},
                statuses=NOT_FINISHED_STATUSES,
            ),
            limit=1,
        )
        work_done = len(run_waiting_to_cancel) == 0

    yield work_done



from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Optional, Union, cast

import dagster._check as check
from dagster._core.definitions import (
    GraphDefinition,
    InputDefinition,
    Node,
    NodeHandle,
    NodeOutput,
    OpDefinition,
)
from dagster._core.definitions.asset_layer import AssetLayer
from dagster._core.definitions.composition import MappedInputPlaceholder
from dagster._core.definitions.dependency import (
    BlockingAssetChecksDependencyDefinition,
    DependencyStructure,
    MultiDependencyDefinition,
    NodeInput,
)
from dagster._core.definitions.executor_definition import ExecutorRequirement
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.repository_definition import RepositoryLoadData
from dagster._core.errors import (
    DagsterExecutionStepNotFoundError,
    DagsterInvariantViolationError,
    DagsterUnmetExecutorRequirementsError,
)
from dagster._core.execution.plan.compute import create_step_outputs
from dagster._core.execution.plan.handle import (
    ResolvedFromDynamicStepHandle,
    StepHandle,
    UnresolvedStepHandle,
)
from dagster._core.execution.plan.inputs import (
    FromConfig,
    FromDefaultValue,
    FromDirectInputValue,
    FromDynamicCollect,
    FromInputManager,
    FromLoadableAsset,
    FromMultipleSources,
    FromMultipleSourcesLoadSingleSource,
    FromPendingDynamicStepOutput,
    FromStepOutput,
    FromUnresolvedStepOutput,
    StepInput,
    StepInputSource,
    StepInputSourceUnion,
    StepInputUnion,
    UnresolvedCollectStepInput,
    UnresolvedMappedStepInput,
)
from dagster._core.execution.plan.instance_concurrency_context import InstanceConcurrencyContext
from dagster._core.execution.plan.outputs import (
    StepOutput,
    StepOutputHandle,
    UnresolvedStepOutputHandle,
)
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.execution.plan.step import (
    ExecutionStep,
    IExecutionStep,
    StepKind,
    UnresolvedCollectExecutionStep,
    UnresolvedMappedExecutionStep,
)
from dagster._core.execution.retries import RetryMode
from dagster._core.instance import DagsterInstance, InstanceRef
from dagster._core.storage.mem_io_manager import mem_io_manager
from dagster._core.system_config.objects import ResolvedRunConfig
from dagster._core.utils import toposort

if TYPE_CHECKING:
    from dagster._core.execution.plan.active import ActiveExecution
    from dagster._core.snap.execution_plan_snapshot import (
        ExecutionPlanSnapshot,
        ExecutionStepInputSnap,
    )


StepHandleTypes = (StepHandle, UnresolvedStepHandle, ResolvedFromDynamicStepHandle)
StepHandleUnion = Union[StepHandle, UnresolvedStepHandle, ResolvedFromDynamicStepHandle]
ExecutionStepUnion = Union[
    ExecutionStep, UnresolvedCollectExecutionStep, UnresolvedMappedExecutionStep
]


class _PlanBuilder:
    """This is the state that is built up during the execution plan build process."""

    def __init__(
        self,
        job_def: JobDefinition,
        resolved_run_config: ResolvedRunConfig,
        step_keys_to_execute: Optional[Sequence[str]],
        known_state: KnownExecutionState,
        instance_ref: Optional[InstanceRef],
        tags: Mapping[str, str],
        repository_load_data: Optional[RepositoryLoadData],
    ):
        self.job_def = check.inst_param(job_def, "job", JobDefinition)
        self.resolved_run_config = check.inst_param(
            resolved_run_config, "resolved_run_config", ResolvedRunConfig
        )
        self.step_keys_to_execute = check.opt_nullable_sequence_param(
            step_keys_to_execute, "step_keys_to_execute", str
        )
        self.known_state = check.inst_param(known_state, "known_state", KnownExecutionState)
        self._instance_ref = instance_ref
        self._tags = check.mapping_param(tags, "tags", key_type=str, value_type=str)
        self.repository_load_data = check.opt_inst_param(
            repository_load_data, "repository_load_data", RepositoryLoadData
        )

        self._steps: dict[str, IExecutionStep] = {}
        self.step_output_map: dict[
            NodeOutput, Union[StepOutputHandle, UnresolvedStepOutputHandle]
        ] = {}
        self._seen_handles: set[StepHandleUnion] = set()

    def add_step(self, step: IExecutionStep) -> None:
        # Keep track of the step keys we've seen so far to ensure we don't add duplicates
        if step.handle in self._seen_handles:
            keys = list(self._steps.keys())
            check.failed(f"Duplicated key {step.key}. Full list seen so far: {keys}.")
        self._seen_handles.add(step.handle)
        self._steps[str(step.node_handle)] = step

    def get_step_by_node_handle(self, handle: NodeHandle) -> IExecutionStep:
        check.inst_param(handle, "handle", NodeHandle)
        return self._steps[str(handle)]

    def build(self) -> "ExecutionPlan":
        """Builds the execution plan."""
        _check_persistent_storage_requirement(
            self.job_def,
            self.resolved_run_config,
        )

        root_inputs: list[
            Union[StepInput, UnresolvedMappedStepInput, UnresolvedCollectStepInput]
        ] = []
        # Recursively build the execution plan starting at the root pipeline
        for input_def in self.job_def.graph.input_defs:
            input_name = input_def.name

            input_source = self.get_root_graph_input_source(
                job_def=self.job_def,
                input_name=input_name,
                input_def=input_def,
            )

            # If an input with dagster_type "Nothing" doesn't have a value
            # we don't create a StepInput
            if input_source is None:
                continue

            root_inputs.append(
                StepInput(
                    name=input_name,
                    dagster_type_key=input_def.dagster_type.key,
                    source=input_source,
                )
            )

        self._build_from_sorted_nodes(
            self.job_def.nodes_in_topological_order,
            self.job_def.dependency_structure,
            parent_step_inputs=root_inputs,
        )

        step_dict = {step.handle: step for step in self._steps.values()}
        step_dict_by_key = {step.key: step for step in self._steps.values()}
        step_handles_to_execute = [step.handle for step in self._steps.values()]

        executable_map, resolvable_map = _compute_step_maps(
            step_dict,
            step_dict_by_key,
            step_handles_to_execute,
            self.known_state,
        )

        executor_name = self.resolved_run_config.execution.execution_engine_name

        plan = ExecutionPlan(
            step_dict,
            executable_map,
            resolvable_map,
            step_handles_to_execute,
            self.known_state,
            _compute_artifacts_persisted(
                step_dict,
                step_dict_by_key,
                step_handles_to_execute,
                self.job_def,
                self.resolved_run_config,
                executable_map,
            ),
            executor_name=executor_name,
            repository_load_data=self.repository_load_data,
        )

        if (
            self.step_keys_to_execute is not None
            # no need to subset if plan already matches request
            and self.step_keys_to_execute != plan.step_keys_to_execute
        ):
            plan = plan.build_subset_plan(
                self.step_keys_to_execute, self.job_def, self.resolved_run_config
            )

        return plan

    def _build_from_sorted_nodes(
        self,
        nodes: Sequence[Node],
        dependency_structure: DependencyStructure,
        parent_handle: Optional[NodeHandle] = None,
        parent_step_inputs: Optional[Sequence[StepInputUnion]] = None,
    ) -> None:
        asset_layer = self.job_def.asset_layer
        step_output_map: dict[NodeOutput, Union[StepOutputHandle, UnresolvedStepOutputHandle]] = {}
        for node in nodes:
            handle = NodeHandle(node.name, parent_handle)

            ### 1. INPUTS
            # Create and add execution plan steps for node inputs
            has_unresolved_input = False
            has_pending_input = False
            step_inputs: list[StepInputUnion] = []
            for input_name, input_def in node.definition.input_dict.items():
                step_input_source = get_step_input_source(
                    self.job_def,
                    node,
                    input_name,
                    input_def,
                    dependency_structure,
                    handle,
                    self.resolved_run_config.ops.get(str(handle)),
                    step_output_map,
                    parent_step_inputs,
                )

                # If an input with dagster_type "Nothing" doesn't have a value
                # we don't create a StepInput
                if step_input_source is None:
                    continue

                if isinstance(
                    step_input_source,
                    (FromPendingDynamicStepOutput, FromUnresolvedStepOutput),
                ):
                    has_unresolved_input = True
                    step_inputs.append(
                        UnresolvedMappedStepInput(
                            name=input_name,
                            dagster_type_key=input_def.dagster_type.key,
                            source=step_input_source,
                        )
                    )
                elif isinstance(step_input_source, FromDynamicCollect):
                    has_pending_input = True
                    step_inputs.append(
                        UnresolvedCollectStepInput(
                            name=input_name,
                            dagster_type_key=input_def.dagster_type.key,
                            source=step_input_source,
                        )
                    )
                else:
                    check.inst_param(
                        step_input_source,
                        "step_input_source",
                        StepInputSource,
                    )
                    step_inputs.append(
                        StepInput(
                            name=input_name,
                            dagster_type_key=input_def.dagster_type.key,
                            source=step_input_source,
                        )
                    )

            ### 2a. COMPUTE FUNCTION
            # Create and add execution plan step for the op compute function
            if isinstance(node.definition, OpDefinition):
                step_outputs = create_step_outputs(
                    node, handle, self.resolved_run_config, asset_layer
                )

                if has_pending_input and has_unresolved_input:
                    check.failed("Can not have pending and unresolved step inputs")

                elif has_unresolved_input:
                    new_step = UnresolvedMappedExecutionStep(
                        handle=UnresolvedStepHandle(node_handle=handle),
                        job_name=self.job_def.name,
                        step_inputs=cast(
                            list[Union[StepInput, UnresolvedMappedStepInput]], step_inputs
                        ),
                        step_outputs=step_outputs,
                        tags=node.tags,
                        pool=node.definition.pool,
                    )
                elif has_pending_input:
                    new_step = UnresolvedCollectExecutionStep(
                        handle=StepHandle(node_handle=handle),
                        job_name=self.job_def.name,
                        step_inputs=cast(
                            list[Union[StepInput, UnresolvedCollectStepInput]], step_inputs
                        ),
                        step_outputs=step_outputs,
                        tags=node.tags,
                        pool=node.definition.pool,
                    )
                else:
                    new_step = ExecutionStep(
                        handle=StepHandle(node_handle=handle),
                        job_name=self.job_def.name,
                        step_inputs=cast(list[StepInput], step_inputs),
                        step_outputs=step_outputs,
                        tags=node.tags,
                        pool=node.definition.pool,
                    )

                self.add_step(new_step)

            ### 2b. RECURSE
            # Recurse over the nodes contained in an instance of GraphDefinition
            elif isinstance(node.definition, GraphDefinition):
                self._build_from_sorted_nodes(
                    node.definition.nodes_in_topological_order,
                    node.definition.dependency_structure,
                    parent_handle=handle,
                    parent_step_inputs=step_inputs,
                )

            else:
                check.invariant(
                    False,
                    f"Unexpected node type {type(node.definition)} encountered during execution"
                    " planning",
                )

            ### 3. OUTPUTS
            # Create output handles for node outputs
            for name, output_def in node.definition.output_dict.items():
                node_output = node.get_output(name)

                # Punch through layers of composition scope to map to the output of the
                # actual compute step
                resolved_output_def, resolved_handle = node.definition.resolve_output_to_origin(
                    output_def.name, handle
                )
                step = self.get_step_by_node_handle(check.not_none(resolved_handle))
                if isinstance(step, (ExecutionStep, UnresolvedCollectExecutionStep)):
                    step_output_handle: Union[StepOutputHandle, UnresolvedStepOutputHandle] = (
                        StepOutputHandle(step.key, resolved_output_def.name)
                    )
                elif isinstance(step, UnresolvedMappedExecutionStep):
                    step_output_handle = UnresolvedStepOutputHandle(
                        step.handle,
                        resolved_output_def.name,
                        step.resolved_by_step_key,
                        step.resolved_by_output_name,
                    )
                else:
                    check.failed(f"Unexpected step type {step}")

                step_output_map[node_output] = step_output_handle

    def get_root_graph_input_source(
        self,
        input_name: str,
        input_def: InputDefinition,
        job_def: JobDefinition,
    ) -> Optional[Union[FromConfig, FromDirectInputValue]]:
        input_values = job_def.input_values
        if input_values and input_name in input_values:
            return FromDirectInputValue(input_name=input_name)

        input_config = self.resolved_run_config.inputs

        if input_config and input_name in input_config:
            return FromConfig(input_name=input_name, node_handle=None)

        if input_def.dagster_type.is_nothing:
            return None

        # Otherwise we throw an error.
        raise DagsterInvariantViolationError(
            f"In top-level graph of {self.job_def.describe_target()}, input {input_name} "
            "must get a value from the inputs section of its configuration."
        )


def get_step_input_source(
    job_def: JobDefinition,
    node: Node,
    input_name: str,
    input_def: InputDefinition,
    dependency_structure: DependencyStructure,
    handle: NodeHandle,
    node_config: Any,
    step_output_map: dict[NodeOutput, Union[StepOutputHandle, UnresolvedStepOutputHandle]],
    parent_step_inputs: Optional[Sequence[StepInputUnion]],
) -> Optional[StepInputSourceUnion]:
    input_handle = node.get_input(input_name)
    input_def = node.definition.input_def_named(input_name)
    asset_layer = job_def.asset_layer

    if (
        # input is unconnected inside the current dependency structure
        not dependency_structure.has_deps(input_handle)
        and
        #  make sure input is unconnected in the outer dependency structure too
        not node.container_maps_input(input_handle.input_name)
    ):
        # can only load from source asset if assets defs are available
        if asset_layer.asset_key_for_input(handle, input_handle.input_name):
            return FromLoadableAsset()
        elif input_def.input_manager_key:
            return FromInputManager(node_handle=handle, input_name=input_name)

    if dependency_structure.has_direct_dep(input_handle):
        node_output_handle = dependency_structure.get_direct_dep(input_handle)
        step_output_handle = step_output_map[node_output_handle]
        if isinstance(step_output_handle, UnresolvedStepOutputHandle):
            return FromUnresolvedStepOutput(
                unresolved_step_output_handle=step_output_handle,
            )

        if node_output_handle.output_def.is_dynamic:
            return FromPendingDynamicStepOutput(
                step_output_handle=step_output_handle,
            )

        return FromStepOutput(
            step_output_handle=step_output_handle,
            fan_in=False,
        )

    if dependency_structure.has_fan_in_deps(input_handle):
        dep_def = dependency_structure.get_dependency_definition(input_handle)
        if isinstance(dep_def, MultiDependencyDefinition):
            return _step_input_source_from_multi_dep_def(
                dependency_structure=dependency_structure,
                input_handle=input_handle,
                step_output_map=step_output_map,
                parent_step_inputs=parent_step_inputs,
                node=node,
                input_name=input_name,
            )
        elif isinstance(dep_def, BlockingAssetChecksDependencyDefinition):
            return _step_input_source_from_blocking_asset_checks_dep_def(
                dep_def=dep_def,
                node_handle=handle,
                dependency_structure=dependency_structure,
                input_handle=input_handle,
                step_output_map=step_output_map,
                parent_step_inputs=parent_step_inputs,
                input_name=input_name,
                asset_layer=asset_layer,
            )
        else:
            check.failed(
                "Expected fan-in deps to correspond to a MultiDependencyDefinition or"
                f" BlockingAssetChecksDependencyDefinition, but was {type(dep_def)}"
            )
    if dependency_structure.has_dynamic_fan_in_dep(input_handle):
        node_output_handle = dependency_structure.get_dynamic_fan_in_dep(input_handle)
        step_output_handle = step_output_map[node_output_handle]
        if isinstance(step_output_handle, UnresolvedStepOutputHandle):
            return FromDynamicCollect(
                source=FromUnresolvedStepOutput(
                    unresolved_step_output_handle=step_output_handle,
                ),
            )
        elif node_output_handle.output_def.is_dynamic:
            return FromDynamicCollect(
                source=FromPendingDynamicStepOutput(
                    step_output_handle=step_output_handle,
                ),
            )

    if node_config and input_name in node_config.inputs:
        return FromConfig(node_handle=handle, input_name=input_name)

    if node.container_maps_input(input_name):
        if parent_step_inputs is None:
            check.failed("unexpected error in composition descent during plan building")

        parent_name = node.container_mapped_input(input_name).graph_input_name
        parent_inputs = {step_input.name: step_input for step_input in parent_step_inputs}
        if parent_name in parent_inputs:
            parent_input = parent_inputs[parent_name]
            return parent_input.source
        # else fall through to Nothing case or raise

    if node.definition.input_has_default(input_name):
        return FromDefaultValue(node_handle=handle, input_name=input_name)

    # At this point we have an input that is not hooked up to
    # the output of another op or provided via run config.

    # We will allow this for "Nothing" type inputs and continue.
    if input_def.dagster_type.is_nothing:
        return None

    # Otherwise we throw an error.
    raise DagsterInvariantViolationError(
        f"In {job_def.describe_target()} {node.describe_node()}, input {input_name} "
        "must get a value either (a) from a dependency or (b) from the "
        "inputs section of its configuration."
    )


def _step_input_source_from_multi_dep_def(
    dependency_structure: DependencyStructure,
    input_handle: NodeInput,
    step_output_map: dict[NodeOutput, Union[StepOutputHandle, UnresolvedStepOutputHandle]],
    parent_step_inputs: Optional[Sequence[StepInputUnion]],
    node: Node,
    input_name: str,
) -> FromMultipleSources:
    sources: list[StepInputSource] = []
    deps = dependency_structure.get_fan_in_deps(input_handle)

    for idx, handle_or_placeholder in enumerate(deps):
        if isinstance(handle_or_placeholder, NodeOutput):
            step_output_handle = step_output_map[handle_or_placeholder]
            if (
                isinstance(step_output_handle, UnresolvedStepOutputHandle)
                or handle_or_placeholder.output_def.is_dynamic
            ):
                check.failed(
                    "Unexpected dynamic output dependency in regular fan in, "
                    "should have been caught at definition time."
                )

            sources.append(
                FromStepOutput(
                    step_output_handle=step_output_handle,
                    fan_in=True,
                )
            )
        else:
            check.invariant(
                handle_or_placeholder is MappedInputPlaceholder,
                f"Expected NodeOutput or MappedInputPlaceholder, got {handle_or_placeholder}",
            )
            if parent_step_inputs is None:
                check.failed("unexpected error in composition descent during plan building")

            parent_name = node.container_mapped_fan_in_input(input_name, idx).graph_input_name
            parent_inputs = {step_input.name: step_input for step_input in parent_step_inputs}
            parent_input = parent_inputs[parent_name]
            source = parent_input.source
            if not isinstance(source, StepInputSource):
                check.failed(f"Unexpected parent mapped input source type {source}")
            sources.append(source)

    return FromMultipleSources(sources=sources)


def _step_input_source_from_blocking_asset_checks_dep_def(
    dep_def: BlockingAssetChecksDependencyDefinition,
    dependency_structure: DependencyStructure,
    input_handle: NodeInput,
    step_output_map: dict[NodeOutput, Union[StepOutputHandle, UnresolvedStepOutputHandle]],
    parent_step_inputs: Optional[Sequence[StepInputUnion]],
    node_handle: NodeHandle,
    input_name: str,
    asset_layer: AssetLayer,
) -> FromMultipleSourcesLoadSingleSource:
    sources: list[StepInputSource] = []
    source_to_load_from: Optional[StepInputSource] = None
    deps = dependency_structure.get_fan_in_deps(input_handle)

    for idx, node_output in enumerate(deps):
        if isinstance(node_output, NodeOutput):
            step_output_handle = step_output_map[node_output]
            if (
                isinstance(step_output_handle, UnresolvedStepOutputHandle)
                or node_output.output_def.is_dynamic
            ):
                check.failed(
                    "Unexpected dynamic output dependency in regular fan in, "
                    "should have been caught at definition time."
                )

            source = FromStepOutput(step_output_handle=step_output_handle, fan_in=True)
            sources.append(source)
            if (
                dep_def.other_dependency is not None
                and dep_def.other_dependency.node == node_output.node_name
                and dep_def.other_dependency.output == node_output.output_name
            ):
                source_to_load_from = source
        else:
            check.invariant(f"Expected NodeOutput, got {node_output}")

    if source_to_load_from is None:
        asset_key_for_input = asset_layer.asset_key_for_input(node_handle, input_handle.input_name)
        if asset_key_for_input:
            source_to_load_from = FromLoadableAsset(node_handle=node_handle, input_name=input_name)
            sources.append(source_to_load_from)
        else:
            check.failed("Unexpected: no sources to load from and no asset key to load from")

    return FromMultipleSourcesLoadSingleSource(
        sources=sources, source_to_load_from=source_to_load_from
    )


class ExecutionPlan(
    NamedTuple(
        "_ExecutionPlan",
        [
            ("step_dict", dict[StepHandleUnion, IExecutionStep]),
            ("executable_map", dict[str, Union[StepHandle, ResolvedFromDynamicStepHandle]]),
            (
                "resolvable_map",
                dict[frozenset[str], Sequence[Union[StepHandle, UnresolvedStepHandle]]],
            ),
            ("step_handles_to_execute", Sequence[StepHandleUnion]),
            ("known_state", KnownExecutionState),
            ("artifacts_persisted", bool),
            ("step_dict_by_key", dict[str, IExecutionStep]),
            ("executor_name", Optional[str]),
            ("repository_load_data", Optional[RepositoryLoadData]),
        ],
    )
):
    def __new__(
        cls,
        step_dict: dict[StepHandleUnion, IExecutionStep],
        executable_map: dict[str, Union[StepHandle, ResolvedFromDynamicStepHandle]],
        resolvable_map: dict[frozenset[str], Sequence[Union[StepHandle, UnresolvedStepHandle]]],
        step_handles_to_execute: Sequence[StepHandleUnion],
        known_state: KnownExecutionState,
        artifacts_persisted: bool = False,
        step_dict_by_key: Optional[dict[str, IExecutionStep]] = None,
        executor_name: Optional[str] = None,
        repository_load_data: Optional[RepositoryLoadData] = None,
    ):
        return super().__new__(
            cls,
            step_dict=check.dict_param(
                step_dict,
                "step_dict",
                key_type=StepHandleTypes,
                value_type=(
                    ExecutionStep,
                    UnresolvedMappedExecutionStep,
                    UnresolvedCollectExecutionStep,
                ),
            ),
            executable_map=executable_map,
            resolvable_map=resolvable_map,
            step_handles_to_execute=check.sequence_param(
                step_handles_to_execute,
                "step_handles_to_execute",
                of_type=(StepHandle, UnresolvedStepHandle, ResolvedFromDynamicStepHandle),
            ),
            known_state=check.inst_param(known_state, "known_state", KnownExecutionState),
            artifacts_persisted=check.bool_param(artifacts_persisted, "artifacts_persisted"),
            step_dict_by_key=(
                {step.key: step for step in step_dict.values()}
                if step_dict_by_key is None
                else check.dict_param(
                    step_dict_by_key,
                    "step_dict_by_key",
                    key_type=str,
                    value_type=(
                        ExecutionStep,
                        UnresolvedMappedExecutionStep,
                        UnresolvedCollectExecutionStep,
                    ),
                )
            ),
            executor_name=check.opt_str_param(executor_name, "executor_name"),
            repository_load_data=check.opt_inst_param(
                repository_load_data, "repository_load_data", RepositoryLoadData
            ),
        )

    @property
    def steps(self) -> Sequence[IExecutionStep]:
        return list(self.step_dict.values())

    @property
    def step_output_versions(self) -> Mapping[StepOutputHandle, str]:
        return self.known_state.step_output_versions if self.known_state else {}

    @property
    def step_keys_to_execute(self) -> Sequence[str]:
        return [handle.to_key() for handle in self.step_handles_to_execute]

    def get_step_output(self, step_output_handle: StepOutputHandle) -> StepOutput:
        check.inst_param(step_output_handle, "step_output_handle", StepOutputHandle)
        return _get_step_output(self.step_dict_by_key, step_output_handle)

    def get_manager_key(
        self,
        step_output_handle: StepOutputHandle,
        job_def: JobDefinition,
    ) -> str:
        return _get_manager_key(self.step_dict_by_key, step_output_handle, job_def)

    def has_step(self, handle: StepHandleUnion) -> bool:
        check.inst_param(handle, "handle", StepHandleTypes)
        return handle in self.step_dict

    def get_step(self, handle: StepHandleUnion) -> IExecutionStep:
        check.inst_param(handle, "handle", StepHandleTypes)
        return self.step_dict[handle]

    def get_step_by_key(self, key: str) -> IExecutionStep:
        check.str_param(key, "key")
        if key not in self.step_dict_by_key:
            check.failed(f"plan has no step with key {key}")
        return self.step_dict_by_key[key]

    def get_executable_step_by_key(self, key: str) -> ExecutionStep:
        step = self.get_step_by_key(key)
        return check.inst(step, ExecutionStep)

    def get_all_steps_in_topo_order(self) -> Sequence[IExecutionStep]:
        return [step for step_level in self.get_all_steps_by_level() for step in step_level]

    def get_all_steps_by_level(self) -> Sequence[Sequence[IExecutionStep]]:
        return [
            [self.get_step_by_key(step_key) for step_key in sorted(step_key_level)]
            for step_key_level in toposort(self.get_all_step_deps())
        ]

    def get_all_step_deps(self) -> Mapping[str, set[str]]:
        deps = {}
        for step in self.step_dict.values():
            if isinstance(step, ExecutionStep):
                deps[step.key] = step.get_execution_dependency_keys()
            elif isinstance(step, (UnresolvedMappedExecutionStep, UnresolvedCollectExecutionStep)):
                deps[step.key] = step.get_all_dependency_keys()
            else:
                check.failed(f"Unexpected execution step type {step}")

        return deps

    def get_steps_to_execute_in_topo_order(self) -> Sequence[ExecutionStep]:
        return [step for step_level in self.get_steps_to_execute_by_level() for step in step_level]

    def get_steps_to_execute_by_level(self) -> Sequence[Sequence[ExecutionStep]]:
        return _get_steps_to_execute_by_level(
            self.step_dict, self.step_dict_by_key, self.step_handles_to_execute, self.executable_map
        )

    def get_executable_step_deps(self) -> Mapping[str, set[str]]:
        return _get_executable_step_deps(
            self.step_dict, self.step_handles_to_execute, self.executable_map
        )

    def resolve(
        self,
        mappings: Mapping[str, Mapping[str, Optional[Sequence[str]]]],
    ) -> Mapping[str, set[str]]:
        """Resolve any dynamic map or collect steps with the resolved dynamic mappings."""
        previous = self.get_executable_step_deps()

        _update_from_resolved_dynamic_outputs(
            self.step_dict,
            self.step_dict_by_key,
            self.executable_map,
            self.resolvable_map,
            self.step_handles_to_execute,
            mappings,
        )

        after = self.get_executable_step_deps()

        return {key: deps for key, deps in after.items() if key not in previous}

    def build_subset_plan(
        self,
        step_keys_to_execute: Sequence[str],
        job_def: JobDefinition,
        resolved_run_config: ResolvedRunConfig,
        step_output_versions: Optional[Mapping[StepOutputHandle, Optional[str]]] = None,
    ) -> "ExecutionPlan":
        check.sequence_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
        step_output_versions = check.opt_mapping_param(
            step_output_versions, "step_output_versions", key_type=StepOutputHandle, value_type=str
        )

        step_handles_to_validate: Sequence[StepHandleUnion] = []
        step_handles_to_validate_set: set[StepHandleUnion] = set()

        # preserve order of step_keys_to_execute since we build the new step_keys_to_execute
        # from iterating step_handles_to_validate
        for key in step_keys_to_execute:
            handle = StepHandle.parse_from_key(key)
            if handle not in step_handles_to_validate_set:
                step_handles_to_validate_set.add(handle)
                step_handles_to_validate.append(handle)

        step_handles_to_execute: list[StepHandleUnion] = []
        bad_keys = []

        for handle in step_handles_to_validate:
            if handle not in self.step_dict:
                # Ok if the entire dynamic step is selected to execute.
                # https://github.com/dagster-io/dagster/issues/8000
                # Note: the assumption here is when the entire dynamic step is selected,
                # the step_keys_to_execute will include both unresolved step (i.e. [?])
                # and all the resolved steps (i.e. [0], ... [n]). Given that at this point
                # we no longer track the parent known state (we don't know what "n" was),
                # solely from the resolved handles, we can't tell if an entire dynamic
                # node is being selected, so the best bet here is to check both unresolved
                # and resolved handles exist. Examples:
                # * `generate_subtasks, subtask[?], subtask[0], subtask[1], subtask[2]` will pass
                # * `generate_subtasks, subtask[0], subtask[1], subtask[2]` will result in 3 bad
                #   keys `subtask[0], subtask[1], subtask[2]`
                if isinstance(handle, ResolvedFromDynamicStepHandle):
                    unresolved_handle = handle.unresolved_form
                    if (
                        unresolved_handle in self.step_dict
                        and unresolved_handle in step_handles_to_validate_set
                    ):
                        continue

                bad_keys.append(handle.to_key())

            # Add the handle to the ready-to-execute list once it's validated
            step_handles_to_execute.append(handle)

        if bad_keys:
            raise DagsterExecutionStepNotFoundError(
                f"Can not build subset plan from unknown step{'s' if len(bad_keys)> 1 else ''}:"
                f" {', '.join(bad_keys)}",
                step_keys=bad_keys,
            )

        executable_map, resolvable_map = _compute_step_maps(
            self.step_dict,
            self.step_dict_by_key,
            step_handles_to_execute,
            self.known_state,
        )

        # If step output versions were provided when constructing the subset plan, add them to the
        # known state.
        known_state = self.known_state
        if len(step_output_versions) > 0:
            known_state = self.known_state._replace(step_output_versions=step_output_versions)

        return ExecutionPlan(
            self.step_dict,
            executable_map,
            resolvable_map,
            step_handles_to_execute,
            known_state,
            _compute_artifacts_persisted(
                self.step_dict,
                self.step_dict_by_key,
                step_handles_to_execute,
                job_def,
                resolved_run_config,
                executable_map,
            ),
            executor_name=self.executor_name,
            repository_load_data=self.repository_load_data,
        )

    def get_version_for_step_output_handle(
        self, step_output_handle: StepOutputHandle
    ) -> Optional[str]:
        return self.step_output_versions.get(step_output_handle)

    def start(
        self,
        retry_mode: RetryMode,
        sort_key_fn: Optional[Callable[[ExecutionStep], float]] = None,
        max_concurrent: Optional[int] = None,
        tag_concurrency_limits: Optional[list[dict[str, Any]]] = None,
        instance_concurrency_context: Optional[InstanceConcurrencyContext] = None,
    ) -> "ActiveExecution":
        from dagster._core.execution.plan.active import ActiveExecution

        return ActiveExecution(
            self,
            retry_mode,
            sort_key_fn,
            max_concurrent,
            tag_concurrency_limits,
            instance_concurrency_context=instance_concurrency_context,
        )

    def step_handle_for_single_step_plans(
        self,
    ) -> Optional[Union[StepHandle, ResolvedFromDynamicStepHandle]]:
        # Temporary hack to isolate single-step plans, which are often the representation of
        # sub-plans in a multiprocessing execution environment.  We want to attribute pipeline
        # events (like resource initialization) that are associated with the execution of these
        # single step sub-plans.  Most likely will be removed with the refactor detailed in
        # https://github.com/dagster-io/dagster/issues/2239
        if len(self.step_handles_to_execute) == 1:
            only_step = self.step_dict[self.step_handles_to_execute[0]]
            if not isinstance(only_step, ExecutionStep):
                return None

            return cast(ExecutionStep, only_step).handle

        return None

    @staticmethod
    def build(
        job_def: JobDefinition,
        resolved_run_config: ResolvedRunConfig,
        step_keys_to_execute: Optional[Sequence[str]] = None,
        known_state: Optional[KnownExecutionState] = None,
        instance_ref: Optional[InstanceRef] = None,
        tags: Optional[Mapping[str, str]] = None,
        repository_load_data: Optional[RepositoryLoadData] = None,
    ) -> "ExecutionPlan":
        """Here we build a new ExecutionPlan from a job definition and the resolved run config.

        To do this, we iterate through the job's nodes in topological order, and hand off the
        execution steps for each node to a companion _PlanBuilder object.

        Once we've processed the entire job, we invoke _PlanBuilder.build() to construct the
        ExecutionPlan object.
        """
        return _PlanBuilder(
            job_def,
            resolved_run_config=resolved_run_config,
            step_keys_to_execute=step_keys_to_execute,
            known_state=known_state or KnownExecutionState(),
            instance_ref=instance_ref,
            tags=tags or {},
            repository_load_data=repository_load_data,
        ).build()

    @staticmethod
    def rebuild_step_input(
        step_input_snap: "ExecutionStepInputSnap",
    ) -> Union["StepInput", "UnresolvedMappedStepInput", "UnresolvedCollectStepInput"]:
        from dagster._core.snap.execution_plan_snapshot import ExecutionStepInputSnap

        check.inst_param(step_input_snap, "step_input_snap", ExecutionStepInputSnap)

        step_input_source = step_input_snap.source

        if isinstance(
            step_input_source,
            (FromPendingDynamicStepOutput, FromUnresolvedStepOutput),
        ):
            return UnresolvedMappedStepInput(
                name=step_input_snap.name,
                dagster_type_key=step_input_snap.dagster_type_key,
                source=step_input_source,
            )
        elif isinstance(step_input_source, FromDynamicCollect):
            return UnresolvedCollectStepInput(
                name=step_input_snap.name,
                dagster_type_key=step_input_snap.dagster_type_key,
                source=step_input_source,
            )
        else:
            check.inst_param(step_input_source, "step_input_source", StepInputSource)
            return StepInput(
                name=step_input_snap.name,
                dagster_type_key=step_input_snap.dagster_type_key,
                source=step_input_snap.source,  # type: ignore  # (possible none)
            )

    @staticmethod
    def rebuild_from_snapshot(
        job_name: str,
        execution_plan_snapshot: "ExecutionPlanSnapshot",
    ) -> "ExecutionPlan":
        if not execution_plan_snapshot.can_reconstruct_plan:
            raise DagsterInvariantViolationError(
                "Tried to reconstruct an old ExecutionPlanSnapshot that was created before"
                " snapshots had enough information to fully reconstruct the ExecutionPlan"
            )

        step_dict: dict[StepHandleUnion, IExecutionStep] = {}
        step_dict_by_key: dict[str, IExecutionStep] = {}

        for step_snap in execution_plan_snapshot.steps:
            input_snaps = step_snap.inputs
            output_snaps = step_snap.outputs

            step_inputs = [
                ExecutionPlan.rebuild_step_input(step_input_snap) for step_input_snap in input_snaps
            ]

            step_outputs = [
                StepOutput(
                    check.not_none(step_output_snap.node_handle),
                    step_output_snap.name,
                    step_output_snap.dagster_type_key,
                    check.not_none(step_output_snap.properties),
                )
                for step_output_snap in output_snaps
            ]

            if step_snap.kind == StepKind.COMPUTE:
                step: IExecutionStep = ExecutionStep(
                    check.inst(
                        step_snap.step_handle,
                        (StepHandle, ResolvedFromDynamicStepHandle),
                    ),
                    job_name,
                    step_inputs,  # type: ignore  # (plain StepInput only)
                    step_outputs,
                    step_snap.tags,
                    step_snap.pool,
                )
            elif step_snap.kind == StepKind.UNRESOLVED_MAPPED:
                step = UnresolvedMappedExecutionStep(
                    check.inst(
                        step_snap.step_handle,
                        UnresolvedStepHandle,
                    ),
                    job_name,
                    step_inputs,  # type: ignore  # (StepInput or UnresolvedMappedStepInput only)
                    step_outputs,
                    step_snap.tags,
                    step_snap.pool,
                )
            elif step_snap.kind == StepKind.UNRESOLVED_COLLECT:
                step = UnresolvedCollectExecutionStep(
                    check.inst(step_snap.step_handle, StepHandle),
                    job_name,
                    step_inputs,  # type: ignore  # (StepInput or UnresolvedCollectStepInput only)
                    step_outputs,
                    step_snap.tags,
                    step_snap.pool,
                )
            else:
                raise Exception(f"Unexpected step kind {step_snap.kind}")

            step_dict[step.handle] = step
            step_dict_by_key[step.key] = step

        step_handles_to_execute = [
            StepHandle.parse_from_key(key) for key in execution_plan_snapshot.step_keys_to_execute
        ]

        executable_map, resolvable_map = _compute_step_maps(
            step_dict,
            step_dict_by_key,
            step_handles_to_execute,
            execution_plan_snapshot.initial_known_state,
        )

        return ExecutionPlan(
            step_dict,
            executable_map,
            resolvable_map,
            step_handles_to_execute,
            # default to empty known execution state if initial was not persisted
            execution_plan_snapshot.initial_known_state or KnownExecutionState(),
            execution_plan_snapshot.artifacts_persisted,
            executor_name=execution_plan_snapshot.executor_name,
            repository_load_data=execution_plan_snapshot.repository_load_data,
        )


def _update_from_resolved_dynamic_outputs(
    step_dict: dict[StepHandleUnion, IExecutionStep],
    step_dict_by_key: dict[str, IExecutionStep],
    executable_map: dict[str, Union[StepHandle, ResolvedFromDynamicStepHandle]],
    resolvable_map: dict[frozenset[str], Sequence[Union[StepHandle, UnresolvedStepHandle]]],
    step_handles_to_execute: Sequence[StepHandleUnion],
    dynamic_mappings: Mapping[str, Mapping[str, Optional[Sequence[str]]]],
) -> None:
    resolved_steps: list[ExecutionStep] = []
    key_sets_to_clear: list[frozenset[str]] = []

    # find entries in the resolvable map whose requirements are now all ready
    for required_keys, unresolved_step_handles in resolvable_map.items():
        if not all(key in dynamic_mappings for key in required_keys):
            continue

        key_sets_to_clear.append(required_keys)

        for unresolved_step_handle in unresolved_step_handles:
            # don't resolve steps we are not executing
            if unresolved_step_handle not in step_handles_to_execute:
                continue

            resolvable_step = step_dict[unresolved_step_handle]

            if isinstance(resolvable_step, UnresolvedMappedExecutionStep):
                resolved_steps += resolvable_step.resolve(dynamic_mappings)
            elif isinstance(resolvable_step, UnresolvedCollectExecutionStep):
                resolved_steps.append(resolvable_step.resolve(dynamic_mappings))

    # update structures
    for step in resolved_steps:
        step_dict[step.handle] = step
        executable_map[step.key] = step.handle
        step_dict_by_key[step.key] = step

    for key_set in key_sets_to_clear:
        del resolvable_map[key_set]


def can_isolate_steps(job_def: JobDefinition) -> bool:
    """Returns true if every output definition in the pipeline uses an IO manager that's not
    the mem_io_manager.

    If true, this indicates that it's OK to execute steps in their own processes, because their
    outputs will be available to other processes.
    """
    output_defs = [
        output_def for node_def in job_def.all_node_defs for output_def in node_def.output_defs
    ]
    for output_def in output_defs:
        if job_def.resource_defs[output_def.io_manager_key] == mem_io_manager:
            return False

    return True


def _check_persistent_storage_requirement(
    job_def: JobDefinition,
    resolved_run_config: ResolvedRunConfig,
) -> None:
    executor_def = job_def.executor_def
    requirements_lst = executor_def.get_requirements(
        resolved_run_config.execution.execution_engine_config
    )
    if ExecutorRequirement.PERSISTENT_OUTPUTS not in requirements_lst:
        return

    if not can_isolate_steps(job_def):
        raise DagsterUnmetExecutorRequirementsError(
            "You have attempted to use an executor that uses multiple processes, but your"
            " job includes op outputs that will not be stored somewhere where other"
            " processes can retrieve them. Please use a persistent IO manager for these outputs."
            ' E.g. with\n   the_graph.to_job(resource_defs={"io_manager": fs_io_manager})'
        )


def should_skip_step(execution_plan: ExecutionPlan, instance: DagsterInstance, run_id: str) -> bool:
    """[INTERNAL] Check if it should skip executing the plan. Primarily used by execution without
    run-level plan process, e.g. Airflow step execution. Note: this only checks one step at a time.

    For each Dagster execution step
    - if none of its inputs come from optional outputs, do not skip
    - if there is at least one input, where none of the upstream steps have yielded an
      output, we should skip the step.
    """
    check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
    check.inst_param(instance, "instance", DagsterInstance)
    check.str_param(run_id, "run_id")

    # only checks one step at a time
    if len(execution_plan.step_keys_to_execute) != 1:
        return False

    step_key = execution_plan.step_keys_to_execute[0]
    optional_source_handles = set()
    step = execution_plan.get_executable_step_by_key(step_key)
    for step_input in step.step_inputs:
        for source_handle in step_input.get_step_output_handle_dependencies():
            if not execution_plan.get_step_output(source_handle).is_required:
                optional_source_handles.add(source_handle)

    # early terminate to avoid unnecessary instance/db calls
    if len(optional_source_handles) == 0:
        # do not skip when all the inputs come from non-optional outputs
        return False

    # find all yielded step outputs
    all_logs = instance.all_logs(run_id)
    yielded_step_output_handles = set()
    for event_record in all_logs:
        if event_record.dagster_event and event_record.dagster_event.is_successful_output:
            yielded_step_output_handles.add(
                event_record.dagster_event.event_specific_data.step_output_handle  # type: ignore
            )

    # If there is at least one of the step's inputs, none of whose upstream steps has
    # yielded an output, we should skip that step.
    for step_input in step.step_inputs:
        missing_source_handles = [
            source_handle
            for source_handle in step_input.get_step_output_handle_dependencies()
            if source_handle in optional_source_handles
            and source_handle not in yielded_step_output_handles
        ]
        if len(missing_source_handles) > 0 and len(missing_source_handles) == len(
            step_input.get_step_output_handle_dependencies()
        ):
            return True

    return False


def _compute_artifacts_persisted(
    step_dict: dict[StepHandleUnion, IExecutionStep],
    step_dict_by_key: dict[str, IExecutionStep],
    step_handles_to_execute: Sequence[StepHandleUnion],
    job_def: JobDefinition,
    resolved_run_config: ResolvedRunConfig,
    executable_map: Mapping[str, Union[StepHandle, ResolvedFromDynamicStepHandle]],
) -> bool:
    """Check if all the border steps of the current run have non-in-memory IO managers for reexecution.

    Border steps: all the steps that don't have upstream steps to execute, i.e. indegree is 0).
    """
    if len(step_dict) == 0:
        return False

    steps_by_level = _get_steps_to_execute_by_level(
        step_dict, step_dict_by_key, step_handles_to_execute, executable_map
    )

    if len(steps_by_level) == 0:
        return False

    for step in steps_by_level[0]:
        # check if all its inputs' upstream step outputs have non-in-memory IO manager configured
        for step_input in step.step_inputs:
            for step_output_handle in step_input.get_step_output_handle_dependencies():
                io_manager_key = _get_manager_key(step_dict_by_key, step_output_handle, job_def)
                manager_def = job_def.resource_defs.get(io_manager_key)
                if (
                    # no IO manager is configured
                    not manager_def
                    # IO manager is non persistent
                    or manager_def == mem_io_manager
                ):
                    return False
    return True


def _get_steps_to_execute_by_level(
    step_dict: Mapping[StepHandleUnion, IExecutionStep],
    step_dict_by_key: Mapping[str, IExecutionStep],
    step_handles_to_execute: Sequence[StepHandleUnion],
    executable_map: Mapping[str, Union[StepHandle, ResolvedFromDynamicStepHandle]],
) -> Sequence[Sequence[ExecutionStep]]:
    return [
        [cast(ExecutionStep, step_dict_by_key[step_key]) for step_key in sorted(step_key_level)]
        for step_key_level in toposort(
            _get_executable_step_deps(step_dict, step_handles_to_execute, executable_map)
        )
    ]


def _get_executable_step_deps(
    step_dict: Mapping[StepHandleUnion, IExecutionStep],
    step_handles_to_execute: Sequence[StepHandleUnion],
    executable_map: Mapping[str, Union[StepHandle, ResolvedFromDynamicStepHandle]],
) -> Mapping[str, set[str]]:
    """Returns:
    Dict[str, Set[str]]: Maps step keys to sets of step keys that they depend on. Includes
        only steps that are included in step_handles_to_execute.
    """
    deps = {}

    # for things transitively downstream of unresolved collect steps
    unresolved_set = set()

    step_keys_to_execute = [handle.to_key() for handle in step_handles_to_execute]

    for key, handle in executable_map.items():
        step = cast(ExecutionStep, step_dict[handle])
        filtered_deps = []
        depends_on_unresolved = False
        for dep in step.get_execution_dependency_keys():
            if dep in executable_map:
                if dep not in unresolved_set:
                    filtered_deps.append(dep)
                else:
                    depends_on_unresolved = True
            elif dep in step_keys_to_execute:
                depends_on_unresolved = True

        if not depends_on_unresolved:
            deps[key] = set(filtered_deps)
        else:
            unresolved_set.add(key)

    return deps


def _get_step_output(step_dict_by_key, step_output_handle: StepOutputHandle) -> StepOutput:
    check.inst_param(step_output_handle, "step_output_handle", StepOutputHandle)
    step = step_dict_by_key[step_output_handle.step_key]
    return step.step_output_named(step_output_handle.output_name)


def _get_manager_key(
    step_dict_by_key: Mapping[str, IExecutionStep],
    step_output_handle: StepOutputHandle,
    pipeline_def: JobDefinition,
) -> str:
    step_output = _get_step_output(step_dict_by_key, step_output_handle)
    node_handle = step_output.node_handle
    output_def = pipeline_def.get_node(node_handle).output_def_named(step_output.name)
    return output_def.io_manager_key


# computes executable_map and resolvable_map and returns them as a tuple. Also
# may modify the passed in step_dict to include resolved step using known_state.
def _compute_step_maps(
    step_dict: dict[StepHandleUnion, IExecutionStep],
    step_dict_by_key: dict[str, IExecutionStep],
    step_handles_to_execute: Sequence[StepHandleUnion],
    known_state: Optional[KnownExecutionState],
) -> tuple[
    dict[str, Union[StepHandle, ResolvedFromDynamicStepHandle]],
    dict[frozenset[str], Sequence[Union[StepHandle, UnresolvedStepHandle]]],
]:
    check.sequence_param(
        step_handles_to_execute,
        "step_handles_to_execute",
        of_type=(StepHandle, UnresolvedStepHandle, ResolvedFromDynamicStepHandle),
    )

    missing_steps = [
        step_handle.to_key()
        for step_handle in step_handles_to_execute
        if step_handle not in step_dict
    ]
    if missing_steps:
        raise DagsterExecutionStepNotFoundError(
            "Execution plan does not contain step{plural}: {steps}".format(
                plural="s" if len(missing_steps) > 1 else "", steps=", ".join(missing_steps)
            ),
            step_keys=missing_steps,
        )

    step_keys_to_execute = [step_handle.to_key() for step_handle in step_handles_to_execute]
    past_mappings = known_state.dynamic_mappings if known_state else {}

    executable_map: dict[str, Union[StepHandle, ResolvedFromDynamicStepHandle]] = {}
    resolvable_map: dict[frozenset[str], list[Union[StepHandle, UnresolvedStepHandle]]] = (
        defaultdict(list)
    )
    for handle in step_handles_to_execute:
        step = step_dict[handle]
        if isinstance(step, ExecutionStep):
            executable_map[step.key] = step.handle
        elif isinstance(step, (UnresolvedMappedExecutionStep, UnresolvedCollectExecutionStep)):
            for key in step.resolved_by_step_keys:
                if key not in step_keys_to_execute and key not in past_mappings:
                    raise DagsterInvariantViolationError(
                        f'Unresolved ExecutionStep "{step.key}" is resolved by "{key}" which is not'
                        " part of the current step selection or known state from parent run."
                    )

            resolvable_map[step.resolved_by_step_keys].append(step.handle)
        else:
            check.invariant(
                step.key in executable_map, "Expect all steps to be executable or resolvable"
            )

    if past_mappings:
        _update_from_resolved_dynamic_outputs(
            step_dict,
            step_dict_by_key,
            executable_map,
            dict(resolvable_map),
            step_handles_to_execute,
            past_mappings,
        )

    return (executable_map, dict(resolvable_map))



from collections.abc import Mapping, Sequence, Set
from functools import cached_property
from typing import Optional, Union

import dagster._check as check
from dagster._config import ConfigFieldSnap, snap_from_field
from dagster._core.definitions import (
    GraphDefinition,
    InputDefinition,
    InputMapping,
    JobDefinition,
    OpDefinition,
    OutputDefinition,
    OutputMapping,
)
from dagster._core.definitions.metadata import (
    MetadataFieldSerializer,
    MetadataValue,
    normalize_metadata,
)
from dagster._core.snap.dep_snapshot import (
    DependencyStructureSnapshot,
    build_dep_structure_snapshot_from_graph_def,
)
from dagster._record import IHaveNew, record, record_custom
from dagster._serdes import whitelist_for_serdes
from dagster._utils.warnings import suppress_dagster_warnings


@whitelist_for_serdes(
    storage_field_names={"metadata": "metadata_entries"},
    field_serializers={"metadata": MetadataFieldSerializer},
    skip_when_empty_fields={"metadata"},
)
@record_custom
class InputDefSnap(IHaveNew):
    name: str
    dagster_type_key: str
    description: Optional[str]
    metadata: Mapping[str, MetadataValue]

    def __new__(
        cls,
        name: str,
        dagster_type_key: str,
        description: Optional[str],
        metadata: Optional[Mapping[str, MetadataValue]] = None,
    ):
        return super().__new__(
            cls,
            name=name,
            dagster_type_key=dagster_type_key,
            description=description,
            metadata=normalize_metadata(
                check.opt_mapping_param(metadata, "metadata", key_type=str), allow_invalid=True
            ),
        )


@whitelist_for_serdes(
    storage_field_names={"metadata": "metadata_entries"},
    field_serializers={"metadata": MetadataFieldSerializer},
    skip_when_empty_fields={"metadata"},
)
@record_custom
class OutputDefSnap(IHaveNew):
    name: str
    dagster_type_key: str
    description: Optional[str]
    is_required: bool
    metadata: Mapping[str, MetadataValue]
    is_dynamic: bool

    def __new__(
        cls,
        name: str,
        dagster_type_key: str,
        description: Optional[str],
        is_required: bool,
        metadata: Optional[Mapping[str, MetadataValue]] = None,
        is_dynamic: bool = False,
    ):
        return super().__new__(
            cls,
            name=name,
            dagster_type_key=dagster_type_key,
            description=description,
            is_required=is_required,
            metadata=normalize_metadata(
                check.opt_mapping_param(metadata, "metadata", key_type=str), allow_invalid=True
            ),
            is_dynamic=is_dynamic,
        )


@whitelist_for_serdes(storage_field_names={"mapped_node_name": "mapped_solid_name"})
@record
class OutputMappingSnap:
    mapped_node_name: str
    mapped_output_name: str
    external_output_name: str


def build_output_mapping_snap(output_mapping: OutputMapping) -> OutputMappingSnap:
    return OutputMappingSnap(
        mapped_node_name=output_mapping.maps_from.node_name,
        mapped_output_name=output_mapping.maps_from.output_name,
        external_output_name=output_mapping.graph_output_name,
    )


@whitelist_for_serdes(storage_field_names={"mapped_node_name": "mapped_solid_name"})
@record
class InputMappingSnap:
    mapped_node_name: str
    mapped_input_name: str
    external_input_name: str


def build_input_mapping_snap(input_mapping: InputMapping) -> InputMappingSnap:
    return InputMappingSnap(
        mapped_node_name=input_mapping.maps_to.node_name,
        mapped_input_name=input_mapping.maps_to.input_name,
        external_input_name=input_mapping.graph_input_name,
    )


def build_input_def_snap(input_def: InputDefinition) -> InputDefSnap:
    check.inst_param(input_def, "input_def", InputDefinition)
    return InputDefSnap(
        name=input_def.name,
        dagster_type_key=input_def.dagster_type.key,
        description=input_def.description,
        metadata=input_def.metadata,
    )


def build_output_def_snap(output_def: OutputDefinition) -> OutputDefSnap:
    check.inst_param(output_def, "output_def", OutputDefinition)
    return OutputDefSnap(
        name=output_def.name,
        dagster_type_key=output_def.dagster_type.key,
        description=output_def.description,
        is_required=output_def.is_required,
        metadata=output_def.metadata,
        is_dynamic=output_def.is_dynamic,
    )


@whitelist_for_serdes(storage_name="CompositeSolidDefSnap", skip_when_empty_fields={"pools"})
@record
class GraphDefSnap:
    name: str
    input_def_snaps: Sequence[InputDefSnap]
    output_def_snaps: Sequence[OutputDefSnap]
    description: Optional[str]
    tags: Mapping[str, str]
    config_field_snap: Optional[ConfigFieldSnap]
    dep_structure_snapshot: DependencyStructureSnapshot
    input_mapping_snaps: Sequence[InputMappingSnap]
    output_mapping_snaps: Sequence[OutputMappingSnap]
    pools: Set[str] = set()

    @cached_property
    def input_def_map(self) -> Mapping[str, InputDefSnap]:
        return {input_def.name: input_def for input_def in self.input_def_snaps}

    @cached_property
    def output_def_map(self) -> Mapping[str, OutputDefSnap]:
        return {output_def.name: output_def for output_def in self.output_def_snaps}

    def get_input_snap(self, name: str) -> InputDefSnap:
        return _get_input_snap(self, name)

    def get_output_snap(self, name: str) -> OutputDefSnap:
        return _get_output_snap(self, name)


@whitelist_for_serdes(storage_name="SolidDefSnap", skip_when_none_fields={"pool"})
@record
class OpDefSnap:
    name: str
    input_def_snaps: Sequence[InputDefSnap]
    output_def_snaps: Sequence[OutputDefSnap]
    description: Optional[str]
    tags: Mapping[str, str]
    required_resource_keys: Sequence[str]
    config_field_snap: Optional[ConfigFieldSnap]
    pool: Optional[str] = None

    @cached_property
    def input_def_map(self) -> Mapping[str, InputDefSnap]:
        return {input_def.name: input_def for input_def in self.input_def_snaps}

    @cached_property
    def output_def_map(self) -> Mapping[str, OutputDefSnap]:
        return {output_def.name: output_def for output_def in self.output_def_snaps}

    def get_input_snap(self, name: str) -> InputDefSnap:
        return _get_input_snap(self, name)

    def get_output_snap(self, name: str) -> OutputDefSnap:
        return _get_output_snap(self, name)


@whitelist_for_serdes(
    storage_name="SolidDefinitionsSnapshot",
    storage_field_names={
        "op_def_snaps": "solid_def_snaps",
        "graph_def_snaps": "composite_solid_def_snaps",
    },
)
@record
class NodeDefsSnapshot:
    op_def_snaps: Sequence[OpDefSnap]
    graph_def_snaps: Sequence[GraphDefSnap]


@suppress_dagster_warnings
def build_node_defs_snapshot(job_def: JobDefinition) -> NodeDefsSnapshot:
    check.inst_param(job_def, "job_def", JobDefinition)
    op_def_snaps = []
    graph_def_snaps = []
    for node_def in job_def.all_node_defs:
        if isinstance(node_def, OpDefinition):
            op_def_snaps.append(build_op_def_snap(node_def))
        elif isinstance(node_def, GraphDefinition):
            graph_def_snaps.append(build_graph_def_snap(node_def))
        else:
            check.failed(f"Unexpected NodeDefinition type {node_def}")

    return NodeDefsSnapshot(
        op_def_snaps=sorted(
            op_def_snaps,
            key=lambda op_def: op_def.name,
        ),
        graph_def_snaps=sorted(
            graph_def_snaps,
            key=lambda graph_def: graph_def.name,
        ),
    )


def build_graph_def_snap(graph_def: GraphDefinition) -> GraphDefSnap:
    check.inst_param(graph_def, "graph_def", GraphDefinition)
    return GraphDefSnap(
        name=graph_def.name,
        input_def_snaps=list(map(build_input_def_snap, graph_def.input_defs)),
        output_def_snaps=list(map(build_output_def_snap, graph_def.output_defs)),
        description=graph_def.description,
        tags=graph_def.tags,
        config_field_snap=(
            snap_from_field("config", graph_def.config_mapping.config_schema.as_field())
            if graph_def.config_mapping
            and graph_def.config_mapping.config_schema
            and graph_def.config_mapping.config_schema.as_field()
            else None
        ),
        dep_structure_snapshot=build_dep_structure_snapshot_from_graph_def(graph_def),
        input_mapping_snaps=list(map(build_input_mapping_snap, graph_def.input_mappings)),
        output_mapping_snaps=list(map(build_output_mapping_snap, graph_def.output_mappings)),
        pools=graph_def.pools,
    )


def build_op_def_snap(op_def: OpDefinition) -> OpDefSnap:
    check.inst_param(op_def, "op_def", OpDefinition)
    return OpDefSnap(
        name=op_def.name,
        input_def_snaps=list(map(build_input_def_snap, op_def.input_defs)),
        output_def_snaps=list(map(build_output_def_snap, op_def.output_defs)),
        description=op_def.description,
        tags=op_def.tags,
        required_resource_keys=sorted(list(op_def.required_resource_keys)),
        config_field_snap=(
            snap_from_field("config", op_def.config_field)  # type: ignore  # (possible none)
            if op_def.has_config_field
            else None
        ),
        pool=op_def.pool,
    )


# shared impl for GraphDefSnap and OpDefSnap
def _get_input_snap(node_def: Union[GraphDefSnap, OpDefSnap], name: str) -> InputDefSnap:
    check.str_param(name, "name")
    inp = node_def.input_def_map.get(name)
    if inp:
        return inp

    check.failed(f"Could not find input {name} in op def {node_def.name}")


# shared impl for GraphDefSnap and OpDefSnap
def _get_output_snap(node_def: Union[GraphDefSnap, OpDefSnap], name: str) -> OutputDefSnap:
    check.str_param(name, "name")
    inp = node_def.output_def_map.get(name)
    if inp:
        return inp

    check.failed(f"Could not find output {name} in node def {node_def.name}")



from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, AbstractSet, Optional  # noqa: UP035

from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.events import AssetKey

if TYPE_CHECKING:
    from dagster._core.definitions.job_definition import JobDefinition
    from dagster._core.definitions.repository_definition import RepositoryDefinition


class IJob(ABC):
    """IJob is a wrapper interface for JobDefinitions to be used as parameters to Dagster's
    core execution APIs.  This enables these execution APIs to operate on both in memory job
    definitions to be executed in the current process (InMemoryJob) as well as definitions that
    can be reconstructed and executed in a different process (ReconstructableJob).
    """

    @abstractmethod
    def get_definition(self) -> "JobDefinition":
        pass

    @abstractmethod
    def get_repository_definition(self) -> Optional["RepositoryDefinition"]:
        pass

    @abstractmethod
    def get_subset(
        self,
        *,
        op_selection: Optional[Iterable[str]] = None,
        asset_selection: Optional[AbstractSet[AssetKey]] = None,
        asset_check_selection: Optional[AbstractSet[AssetCheckKey]] = None,
    ) -> "IJob":
        pass

    @property
    @abstractmethod
    def op_selection(self) -> Optional[AbstractSet[str]]:
        pass

    @property
    @abstractmethod
    def asset_selection(self) -> Optional[AbstractSet[AssetKey]]:
        pass

    @property
    @abstractmethod
    def asset_check_selection(self) -> Optional[AbstractSet[AssetCheckKey]]:
        pass

    @property
    def resolved_op_selection(self) -> Optional[AbstractSet[str]]:
        return set(self.op_selection) if self.op_selection else None


class InMemoryJob(IJob):
    def __init__(
        self,
        job_def: "JobDefinition",
    ):
        self._job_def = job_def

    def get_definition(self) -> "JobDefinition":
        return self._job_def

    def get_repository_definition(self) -> Optional["RepositoryDefinition"]:
        return None

    def get_subset(
        self,
        *,
        op_selection: Optional[Iterable[str]] = None,
        asset_selection: Optional[AbstractSet[AssetKey]] = None,
        asset_check_selection: Optional[AbstractSet[AssetCheckKey]] = None,
    ) -> "InMemoryJob":
        op_selection = set(op_selection) if op_selection else None
        return InMemoryJob(
            self._job_def.get_subset(
                op_selection=op_selection,
                asset_selection=asset_selection,
                asset_check_selection=asset_check_selection,
            )
        )

    @property
    def op_selection(self) -> Optional[AbstractSet[str]]:
        return self._job_def.op_selection

    @property
    def asset_selection(self) -> Optional[AbstractSet[AssetKey]]:
        return self._job_def.asset_selection

    @property
    def asset_check_selection(self) -> Optional[AbstractSet[AssetCheckKey]]:
        return self._job_def.asset_check_selection



# alembic dynamically populates the alembic.context module

from alembic import context
from dagster._core.storage.event_log import SqlEventLogStorageMetadata
from dagster._core.storage.runs import SqlRunStorage
from dagster._core.storage.schedules import SqlScheduleStorage
from dagster._core.storage.sql import run_migrations_offline, run_migrations_online

config = context.config

target_metadata = [SqlEventLogStorageMetadata, SqlRunStorage, SqlScheduleStorage]

if context.is_offline_mode():
    run_migrations_offline(context, config, target_metadata)  # type: ignore  # (bad stubs)
else:
    run_migrations_online(context, config, target_metadata)  # type: ignore  # (bad stubs)
